/** Group specific routers for the email service
 * in expressjs
 */
import express, { Router } from "express";
import { z } from "zod";

import { signGroupInviteToken } from "@snip/auth/tokens/groupInvite";
import service from "@snip/database";
import { NotFoundError } from "@snip/database/errors";
import { GroupData, GroupMemberData, UserData } from "@snip/database/types";

import { authMiddleWare } from "@/common/auth";
import { asyncHandler } from "@/common/errorHandler";
import { GroupInviteEmail } from "@/email/email";
import { sendEmail } from "@/email/send";

export const groupRouter: Router = express.Router();

groupRouter.post(
    "/inviteUser",
    authMiddleWare,
    express.json(),
    asyncHandler(async (req, res) => {
        const bodySchema = z.object({
            //User to invite
            email: z.string().email(),

            //Group id
            id: z.number().int(),
        });

        const parsedBody = bodySchema.safeParse(req.body);

        if (!parsedBody.success) {
            res.status(400).json({
                error: "Invalid request body",
                details: "Valid Email required",
            });
            return;
        }

        const emailAdrr = parsedBody.data.email;
        const group_id = parsedBody.data.id;
        const user_id = req.session.user.id;

        // Check if user is allowed to invite
        // this also checks if the user is in group
        // i.e. in group and is owner/moderator

        const isAllowedToEdit = await service.group.isAllowedToEdit(
            user_id,
            group_id,
        );

        if (!isAllowedToEdit) {
            res.status(400).json({
                error: "Not allowed to invite",
            });
            return;
        }

        // Check if group exists
        let group: GroupData;
        try {
            group = await service.group.getById(group_id);
        } catch (error) {
            if (error instanceof NotFoundError) {
                res.status(400).json({
                    error: "Group not found",
                });
                return;
            }
            throw error;
        }

        // Check if user exists and is not already in group
        let invitedUser: UserData;
        let potentialMember: GroupMemberData;
        try {
            invitedUser = await service.user.getByEmail(emailAdrr);
            potentialMember = await service.group
                .getMember(invitedUser.id, group_id)
                .catch((error) => {
                    if (error instanceof NotFoundError) {
                        return null;
                    }
                    throw error;
                });
        } catch (error) {
            if (error instanceof NotFoundError) {
                res.status(400).json({
                    error: "User not found",
                });
                return;
            }
            throw error;
        }
        if (potentialMember && potentialMember.role !== "invited") {
            res.status(400).json({
                error: "User already in group",
            });
            return;
        }

        // Check if invite is still valid
        // Allow to resend every 2hours
        if (
            potentialMember &&
            potentialMember.role === "invited" &&
            Date.now() - potentialMember.joined_at.getTime() <
                2 * 60 * 60 * 1000
        ) {
            res.status(400).json({
                error: "Invite is still pending for this user!",
                details: `You may try again in ${
                    (2 * 60 * 60 * 1000 -
                        (Date.now() - potentialMember.joined_at.getTime())) /
                    1000
                }s`,
            });
            return;
        }

        // Add db entry
        const member = await service.group.addMember(
            invitedUser.id,
            group_id,
            "invited",
        );

        // Get user for email information i.e the inviter
        const user = await service.user.getById(user_id);

        // Generate token and send email
        const token = await signGroupInviteToken(
            invitedUser,
            member,
            user,
            group,
        );
        const email = new GroupInviteEmail(
            invitedUser.email,
            group.name,
            group.id,
            token,
            user.email,
        );
        try {
            await sendEmail(email);
            res.json({
                member,
                success: true,
            });
        } catch (error) {
            res.status(500).json({
                error: "Failed to send email",
                details: (error as Error).message,
            });
        }
    }),
);
