"use client";

import React, { ChangeEvent, useEffect, useRef, useState } from "react";

import { FONT_URLS } from "@snip/database/fonts/internal";
import { DoodleSnip } from "@snip/snips/general/doodle";
import { TextSnip } from "@snip/snips/general/text";

import { useUserProfile } from "../context";
import { lorem } from "./data";

import { OverlayWorker } from "components/editor/worker/overlayWorker";
import { FormInfo } from "components/form/alert";
import { Input } from "components/form/inputs/base";
import { ColorInput, MultiColorInput } from "components/form/inputs/color";
import {
    JoinedRangeNumberInput,
    NumberInput,
} from "components/form/inputs/number";
import { DropdownInput } from "components/form/inputs/option";
import { Item } from "components/form/items";

import styles from "./settings.module.scss";

export default function Settings() {
    return (
        <>
            <h2>Tool settings</h2>
            <div className="d-flex p-1 mb-2">
                The tool settings are bound to the user and will be used on any
                device you are logged in. These settings are the default values
                for the tools in the viewer. For them to take effect you need to
                reload the viewer.
            </div>
            <Item
                title="Doodle tool"
                description="The doodle tool is used to draw on the page in freehand form. You can change the size, color, and smoothing of the pen."
            >
                <DoodleConfig />
                <TestDoodle />
            </Item>
            <Item
                title="Text tool"
                description="The text tool is used to add text to the page. You can change the font, size, color, line height, and line wrap."
            >
                <FontConfig />
                <FontPreview />
            </Item>
            <FormInfo style={{ marginInline: "3rem" }}>
                We might add more tool configuration options here in the future.
                If have any suggestions, please let us know.
            </FormInfo>
        </>
    );
}

function DoodleConfig() {
    const { config, updateUser } = useUserProfile();

    return (
        <>
            <JoinedRangeNumberInput
                value={config?.pen_size + ""}
                label="Default pen size"
                name="pen_size"
                onInput={(e: ChangeEvent<HTMLInputElement>) => {
                    const value = parseFloat(e.target.value);
                    updateUser(null, { id: config.id, pen_size: value });
                }}
                min={1}
                max={25}
                step={1}
                onReset={() => {
                    updateUser(null, { id: config.id, pen_size: 4 });
                }}
            />

            <MultiColorInput
                label="Default Colors"
                colorList={config.pen_colors}
                currentColorIdx={config.pen_colors.indexOf(config.pen_color)}
                setCurrentColorIdx={(idx: number) => {
                    updateUser(null, {
                        id: config.id,
                        pen_color: config.pen_colors[idx],
                    });
                }}
                setColorList={(colors: string[]) => {
                    updateUser(null, {
                        id: config.id,
                        pen_colors: colors,
                    });
                }}
                onReset={() => {
                    updateUser(null, {
                        id: config.id,
                        pen_colors: [
                            "#000000",
                            "#ff0000",
                            "#00ff00",
                            "#0000ff",
                        ],
                        pen_color: "#000000",
                    });
                }}
                showHex
            />

            <JoinedRangeNumberInput
                value={config?.pen_smoothing.toFixed(2)}
                label="Default pen smoothing"
                name="pen_smoothing"
                onInput={(e: ChangeEvent<HTMLInputElement>) => {
                    const value = parseFloat(e.target.value);
                    updateUser(null, {
                        id: config.id,
                        pen_smoothing: value,
                    });
                }}
                min={0}
                max={1}
                step={0.01}
                onReset={() => {
                    updateUser(null, {
                        id: config.id,
                        pen_smoothing: 0.0,
                    });
                }}
            />
        </>
    );
}

function TestDoodle() {
    const [worker, setWorker] = useState<OverlayWorker>();
    const containerRef = useRef<HTMLDivElement>();
    const { config } = useUserProfile();

    //Attach event listener
    useEffect(() => {
        if (!worker) return;
        const canvas = containerRef.current;

        //x,y,line/move
        const stack: [DoodleSnip, number][] = [];

        function addEdge(e: MouseEvent, n: boolean = false) {
            const xy = worker.canvasCoords_2_dbCoords_het([
                e.offsetX,
                e.offsetY,
            ]) as [number, number];

            if (n) {
                stack.push([
                    new DoodleSnip({
                        edges: [],
                        colour: config?.pen_color || "#000000",
                        lineWidth: config?.pen_size || 4,
                        smoothing: config?.pen_smoothing || 0,
                        book_id: -1,
                    }),
                    performance.now(),
                ]);
            }

            let snip = stack.at(-1)?.[0];
            if (!snip) {
                snip = new DoodleSnip({
                    edges: [],
                    colour: config?.pen_color || "#000000",
                    lineWidth: config?.pen_size || 4,
                    smoothing: config?.pen_smoothing || 0,
                    book_id: -1,
                });
            }

            snip.add_edge([...xy]);
        }

        const render = (ctx: CanvasRenderingContext2D) => {
            const to_remove = [];
            for (let i = 0; i < stack.length; i++) {
                const [snip, time] = stack[i];
                if (performance.now() - time > 5000) {
                    to_remove.push(i);
                } else {
                    snip.render(ctx);
                }
            }

            //remove
            for (let i = to_remove.length - 1; i >= 0; i--) {
                stack.splice(to_remove[i], 1);
            }
        };
        worker.add(render);

        /* -------------------------------------------------------------------------- */
        /*                                event handler                               */
        /* -------------------------------------------------------------------------- */

        const onMouseDown = (e: MouseEvent) => {
            addEdge(e, true);

            document.addEventListener("mousemove", onMouseMove);
            document.addEventListener("mouseup", onMouseUp);
            canvas.addEventListener("mouseleave", onMouseUp);
        };
        const onMouseMove = (e: MouseEvent) => {
            //return not mouse down
            if (e.buttons !== 1) return;
            addEdge(e);
        };
        const onMouseUp = () => {
            document.removeEventListener("mouseup", onMouseUp);
            document.removeEventListener("mousemove", onMouseMove);
        };

        canvas.addEventListener("mousedown", onMouseDown);

        return () => {
            worker.remove(render);
            canvas.removeEventListener("mousedown", onMouseDown);
            canvas.removeEventListener("mousemove", onMouseMove);
            canvas.removeEventListener("mouseup", onMouseUp);
        };
    }, [config?.pen_color, config?.pen_size, config?.pen_smoothing, worker]);

    return (
        <div ref={containerRef} className={styles.testDoodle}>
            <Input
                as="div"
                label={
                    <span>Test your configuration here (width=page width)</span>
                }
            >
                <Preview worker={worker} setWorker={setWorker} animation />
            </Input>
        </div>
    );
}

function Preview({
    worker,
    setWorker,
    animation = false,
}: {
    animation?: boolean;
    worker: OverlayWorker;
    setWorker: (worker: OverlayWorker) => void;
}) {
    const canvasRef = useRef<HTMLCanvasElement>();

    // Setup worker
    useEffect(() => {
        if (worker || !canvasRef.current) return;
        if (canvasRef.current) {
            const worker = new OverlayWorker(undefined);
            worker.setupCanvas(canvasRef.current);
            setWorker(worker);

            return () => {
                worker.terminate();
            };
        }
    }, [canvasRef, setWorker, worker]);

    //Setup observer
    useEffect(() => {
        if (!worker) return;

        const render = () => {
            worker.rerender();
            if (animation) window.requestAnimationFrame(render);
        };
        window.requestAnimationFrame(render);

        /* -------------------------------------------------------------------------- */
        /*                               resize observer                              */
        /* -------------------------------------------------------------------------- */
        const resizeObserver = new ResizeObserver(() => {
            const width = canvasRef.current.clientWidth;
            const height = canvasRef.current.clientHeight;
            worker.resize(width, height, 1, true);
        });
        resizeObserver.observe(canvasRef.current);

        return () => {
            resizeObserver.disconnect();
        };
    }, [animation, worker]);

    return (
        <div
            style={{
                width: "100%",
                height: "6rem",
            }}
        >
            <canvas ref={canvasRef}></canvas>
        </div>
    );
}

function FontConfig() {
    const { config, updateUser } = useUserProfile();

    return (
        <>
            <TextFontInput
                value={config?.text_font}
                onSelect={(key) => {
                    updateUser(null, { id: config.id, text_font: key });
                }}
                onReset={() => {
                    updateUser(null, { id: config.id, text_font: "Arial" });
                }}
            />
            <ColorInput
                color={config?.text_fontColor}
                setColor={(color) => {
                    updateUser(null, { id: config.id, text_fontColor: color });
                }}
                showHex
                label="Font color"
                name="text_fontColor"
                onInput={(e: ChangeEvent<HTMLInputElement>) => {
                    const value = e.target.value;
                    updateUser(null, {
                        id: config.id,
                        text_fontColor: value,
                    });
                }}
                onReset={() => {
                    updateUser(null, {
                        id: config.id,
                        text_fontColor: "#000000",
                    });
                }}
            />
            <NumberInput
                value={config?.text_fontSize.toFixed(0)}
                label="Font size"
                name="text_fontSize"
                onInput={(e: ChangeEvent<HTMLInputElement>) => {
                    const value = parseFloat(e.target.value);
                    updateUser(null, {
                        id: config.id,
                        text_fontSize: value,
                    });
                }}
                min={1}
                max={25}
                step={1}
                onReset={() => {
                    updateUser(null, { id: config.id, text_fontSize: 12 });
                }}
            />
            <NumberInput
                value={config?.text_lineHeight.toFixed(2)}
                label="Line height"
                name="text_lineHeight"
                onInput={(e: ChangeEvent<HTMLInputElement>) => {
                    const value = parseFloat(e.target.value);
                    updateUser(null, {
                        id: config.id,
                        text_lineHeight: value,
                    });
                }}
                min={0.1}
                max={3}
                step={0.01}
                onReset={() => {
                    updateUser(null, {
                        id: config.id,
                        text_lineHeight: 1.25,
                    });
                }}
            />
            <NumberInput
                value={config?.text_lineWrap.toFixed(0)}
                label="Line wrap"
                name="text_lineWrap"
                onInput={(e: ChangeEvent<HTMLInputElement>) => {
                    const value = parseFloat(e.target.value);
                    updateUser(null, {
                        id: config.id,
                        text_lineWrap: value,
                    });
                }}
                min={100}
                max={1000}
                step={1}
                onReset={() => {
                    updateUser(null, { id: config.id, text_lineWrap: 400 });
                }}
            />
        </>
    );
}

function FontPreview() {
    const [worker, setWorker] = useState<OverlayWorker>();
    const { config } = useUserProfile();

    useEffect(() => {
        if (!config || !worker) return;

        const text = new TextSnip({
            text: lorem,
            fontFamily: config.text_font,
            fontSize: config.text_fontSize,
            colour: config.text_fontColor,
            lineHeight: config.text_lineHeight,
            lineWrap: config.text_lineWrap,
            book_id: -1,
        });

        const render = (ctx: CanvasRenderingContext2D) => {
            //Rotate by 94 degrees
            ctx.save();
            ctx.translate(700 - text.width / 2, 10);
            text.render(ctx);
            ctx.restore();
        };

        worker.add(render);

        return () => {
            worker.remove(render);
        };
    }, [config, worker]);

    return (
        <div className={styles.testDoodle}>
            <Input
                as="div"
                label="Example (width=page width)"
                className={styles.testDoodle}
            >
                <div
                    style={{
                        width: "100%",
                        height: "20rem",
                    }}
                >
                    <Preview worker={worker} setWorker={setWorker} />
                </div>
            </Input>
        </div>
    );
}

function TextFontInput({
    value,
    onSelect = () => {},
    onReset = () => {},
}: {
    value: string;
    onSelect: (key: string, e: React.SyntheticEvent<unknown>) => void;
    onReset: () => void;
}) {
    return (
        <DropdownInput
            value={value}
            label="Default font family"
            options={FONT_URLS.map(([, name]) => {
                return {
                    label: <span style={{ fontFamily: name }}>{name}</span>,
                    eventKey: name,
                };
            })}
            onSelect={onSelect}
            onReset={onReset}
        />
    );
}
