/**
 * @jest-environment node
 */
import { createMockRequest } from "__mocks__/request";
import { POST, GET } from "../route";
import { setSessionType } from "__mocks__/mockConfig";

/** Mock the database connection
 * so that we can test the API
 */
let force_error = false;
jest.mock("@snip/database", () => {
    return {
        book: {
            insert: jest.fn(async (data) => {
                if (force_error) {
                    throw Error("Mock error");
                }
                return {
                    ...data,
                    id: -10,
                };
            }),
            getByUserId: jest.fn(async (user_id) => {
                if (force_error) {
                    throw Error("Mock error");
                }
                return [
                    {
                        id: -10,
                        title: "Test",
                        owner_user_id: 42,
                        last_updated: new Date().toISOString(),
                    },
                ];
            }),
        },
    };
});

describe("POST api/books/ (create)", () => {
    describe("should gracefully handle errors", () => {
        it("return 400 on invalid query", async () => {
            const req = await createMockRequest({
                method: "POST",
                content_type: "application/json",
                body: "asdasdaasd!",
            });
            const res = await POST(req, {});
            expect(res.status).toBe(400);
            expect(await res.json()).toEqual({
                error: "Bad Request",
                details: "Expected object, received string",
            });
        });

        it("return 400 on missing title", async () => {
            const req = await createMockRequest({
                method: "POST",
                content_type: "application/json",
                body: {},
            });
            const res = await POST(req, {});
            expect(res.status).toBe(400);
            expect(await res.json()).toEqual({
                error: "Bad Request",
                details: "The book title is required",
            });
        });

        it("return 400 on invalid title", async () => {
            const req = await createMockRequest({
                method: "POST",
                content_type: "application/json",
                body: { title: "a" },
            });
            const res = await POST(req, {});
            expect(res.status).toBe(400);
            expect(await res.json()).toEqual({
                error: "Bad Request",
                details: "The book title must be at least 3 characters",
            });

            const req2 = await createMockRequest({
                method: "POST",
                content_type: "application/json",
                body: { title: "a".repeat(257) },
            });
            const res2 = await POST(req2, {});
            expect(res2.status).toBe(400);
            expect(await res2.json()).toEqual({
                error: "Bad Request",
                details: "The book title must be at most 256 characters",
            });
        });
    });

    it("should create a book", async () => {
        const req = await createMockRequest({
            method: "POST",
            content_type: "application/json",
            body: { title: "Test" },
        });
        const res = await POST(req, {});
        expect(res.status).toBe(200);
        expect(await res.json()).toEqual({
            id: -10,
            title: "Test",
            owner_user_id: 42,
            last_updated: expect.any(String),
        });
    });

    it("should create a book with comment", async () => {
        const req = await createMockRequest({
            method: "POST",
            content_type: "application/json",
            body: { title: "Test", comment: "Test comment" },
        });
        const res = await POST(req, {});
        expect(res.status).toBe(200);
        expect(await res.json()).toEqual({
            id: -10,
            title: "Test",
            comment: "Test comment",
            owner_user_id: 42,
            last_updated: expect.any(String),
        });
    });
});

describe("GET api/books/ (get)", () => {
    it("should get books", async () => {
        force_error = false;
        const req = await createMockRequest({
            method: "GET",
        });
        const res = await GET(req, {});
        expect(res.status).toBe(200);
        expect(await res.json()).toEqual([
            {
                id: -10,
                title: "Test",
                owner_user_id: 42,
                last_updated: expect.any(String),
            },
        ]);

        // Test with api session
        setSessionType("api_token");
        const res2 = await GET(req, {});
        expect(res2.status).toBe(401);
        setSessionType("ui_token");
        const res3 = await GET(req, {});
        expect(res3.status).toBe(401);
    });
});
