"use client";
import { usePathname, useSearchParams } from "next/navigation";
import {
    createContext,
    Dispatch,
    MutableRefObject,
    SetStateAction,
    useCallback,
    useContext,
    useEffect,
    useMemo,
    useRef,
    useState,
} from "react";

import { ID } from "@snip/database/types";

import { usePagesContext } from "./socket/usePagesContext";
import { WorkerContextProvider } from "./viewer/useWorkerContext";

interface ViewerContext extends CanvasRefs {
    // The active page ids
    activePageIds: ID[];
    activePageIdsRef: MutableRefObject<ID[]>;
    setActivePageIds: Dispatch<SetStateAction<ID[]>>;

    // The active snip id (if in snip view mode)
    activeSnipId: ID;
    activeSnipIdRef: MutableRefObject<ID>;
    setActiveSnipId: Dispatch<SetStateAction<ID>>;

    // The active book id
    book_id: ID;
}

export type EleMap<T> = Map<number, T>;

interface CanvasRefs {
    // refs for the layout of the pages
    draggableRef: MutableRefObject<EleMap<HTMLDivElement>>;
    wrapperRef: MutableRefObject<EleMap<HTMLDivElement>>;
    containerRef: MutableRefObject<EleMap<HTMLDivElement>>;
}

const ViewerContext = createContext<ViewerContext>(undefined);

export function ViewerContextProvider({
    children,
    book_id,
    initialActivePageIds,
}: {
    children: React.ReactNode;
    book_id: ID;
    initialActivePageIds: ID[];
}) {
    const [activePageIds, setActivePageIds] =
        useActivePageIds(initialActivePageIds);
    const activePageIdsRef = useRef(activePageIds);
    const [activeSnipId, setActiveSnipId] = useState<ID>(null);
    const activeSnipIdRef = useRef(activeSnipId);

    const canvasRefs = useCanvasRefs();

    const ret: ViewerContext = useMemo(() => {
        return {
            book_id,
            activePageIds,
            setActivePageIds,
            activePageIdsRef,
            activeSnipId,
            setActiveSnipId,
            activeSnipIdRef,
            ...canvasRefs,
        };
    }, [book_id, activePageIds, setActivePageIds, activeSnipId, canvasRefs]);

    return (
        <ViewerContext.Provider value={ret}>
            <WorkerContextProvider n={activePageIds.length}>
                {children}
            </WorkerContextProvider>
        </ViewerContext.Provider>
    );
}

export function useViewerContext() {
    const context = useContext(ViewerContext);
    if (!context) {
        throw new Error(
            "useViewerContext must be used within a ViewerContextProvider",
        );
    }
    return context;
}

/** This wrapper handles the searchParams update for
 * the active page ids and allows to select a
 * page by searchParams
 */
function useActivePageIds(
    initialActivePageIds: ID[],
): [ID[], Dispatch<SetStateAction<ID[]>>] {
    const pathname = usePathname();
    const searchParams = useSearchParams();

    const { pages } = usePagesContext();

    const [activePageIds, setActivePageIds] =
        useState<ID[]>(initialActivePageIds);

    const setSearchParams = useCallback(
        (page_numbers: number[]) => {
            const search = new URLSearchParams(searchParams.toString());

            //Remove all page params
            search.delete("page");
            for (const no of page_numbers) {
                search.append("page", (no + 1).toString());
            }
            window.history.replaceState(
                {},
                "",
                `${pathname}?${search.toString()}`,
            );
        },
        [pathname, searchParams],
    );

    useEffect(() => {
        if (!pages || pages.size < 1) return;
        setSearchParams(activePageIds.map((id) => pages.get(id).page_number));
    }, [activePageIds, pages, setSearchParams]);

    return [activePageIds, setActivePageIds];
}

/**
 * Sets a reference to an element in a mutable ref object at the specified index.
 *
 * @template T - The type of the element.
 * @param ref - The mutable ref object to set the reference in.
 * @param idx - The index at which to set the reference.
 * @returns A function that sets the reference to the element.
 */
export function setRef<T>(ref: MutableRefObject<EleMap<T>>, idx: number) {
    return (element: T) => {
        if (!element) return;
        if (!ref.current) ref.current = new Map<number, T>();
        ref.current.set(idx, element);
    };
}

/** The canvas refs allow
 * to access the canvases of the pages
 * (if they are in the dom)
 */
function useCanvasRefs(): CanvasRefs {
    /** Define the references as arrays as we
     * allow multiple pages to be rendered
     */

    const draggableRef = useRef<EleMap<HTMLDivElement>>(null),
        wrapperRef = useRef<EleMap<HTMLDivElement>>(null),
        containerRef = useRef<EleMap<HTMLDivElement>>(null);

    return {
        draggableRef,
        wrapperRef,
        containerRef,
    };
}
