"use client";
import { createContext, useContext, useEffect, useRef, useState } from "react";

import { useSocketContext } from "../useSocketContext";

import { OverlayWorker } from "components/editor/worker/overlayWorker";
import { PageWorker } from "components/editor/worker/pageWorker";

export type EleMap<T> = Map<number, T>;

interface WorkerContext {
    worker: EleMap<PageWorker>;
    overlayWorker: EleMap<OverlayWorker>;
}

const WorkerContext = createContext<WorkerContext>({
    worker: new Map(),
    overlayWorker: new Map(),
});

export function WorkerContextProvider({
    n,
    children,
}: {
    n: number;
    children: React.ReactNode;
}) {
    const { socket } = useSocketContext();

    const [worker, setWorker] = useState<EleMap<PageWorker>>(new Map());
    const [overlayWorker, setOverlayWorker] = useState<EleMap<OverlayWorker>>(
        new Map(),
    );

    /* Scaling and create the workers dynamically */
    useEffect(() => {
        if (!socket) return;
        setWorker((w) => {
            //Scale up
            if (w.size < n) {
                for (let i = w.size; i < n; i++) {
                    w.set(w.size, new PageWorker());
                }
                return new Map(w);
            }
            //Scale down
            else if (w.size > n) {
                for (let i = w.size - 1; i >= n; i--) {
                    const worker = w.get(i);
                    if (worker) {
                        worker.terminate().then(() => {
                            w.delete(i);
                        });
                    }
                }
                return new Map(w);
            }
            return w;
        });

        setOverlayWorker((o) => {
            if (o.size < n) {
                for (let i = o.size; i < n; i++) {
                    o.set(o.size, new OverlayWorker(socket));
                }
                return new Map(o);
            } else if (o.size > n) {
                for (let i = o.size - 1; i >= n; i--) {
                    const worker = o.get(i);
                    if (worker) {
                        worker.terminate();
                        o.delete(i);
                    }
                }
                return new Map(o);
            }
            return o;
        });
    }, [n, socket]);

    useEffect(() => {
        setWorker((p) => p);
        setOverlayWorker((o) => o);
        return () => {
            setWorker((w) => {
                w.forEach((worker, i) => {
                    worker.terminate();
                    w.delete(i);
                });
                return new Map();
            });
            setOverlayWorker((o) => {
                o.forEach((worker, i) => {
                    worker.terminate();
                    o.delete(i);
                });
                return new Map();
            });
        };
    }, []);

    return (
        <WorkerContext.Provider value={{ worker, overlayWorker }}>
            {children}
        </WorkerContext.Provider>
    );
}

export function useWorkerContext() {
    const context = useContext(WorkerContext);
    if (!context) {
        throw new Error(
            "useWorkerContext must be used within a WorkerContextProvider",
        );
    }
    return context;
}

export function useWorker(idx: number): {
    worker: PageWorker | undefined;
    overlayWorker: OverlayWorker | undefined;
} {
    const { worker, overlayWorker } = useWorkerContext();
    return {
        worker: worker?.get(idx),
        overlayWorker: overlayWorker?.get(idx),
    };
}
