import { useState } from "react";
import { TbPhotoEdit } from "react-icons/tb";

import DragMove from "../common/drag";
import { Help, HelpDescription, HelpHeader } from "../common/help";
import { Overlay, OverlayContent, useSnipOverlay } from "../common/overlay";
import { ItemSmall, Submenu } from "../common/submenu";
import { Toggle } from "../common/toggle";
import { usePlacementContext } from "../placement/context";
import { ToolComponents, ToolSubmenuProps, ToolToggleProps } from "../types";
import { ImageEditorContextProvider, useImageEditorContext } from "./context";
import {
    CutHandlersOverlayOutline,
    ResizeHandlerOverlayOutline,
} from "./handlers";

import { ResetButton } from "components/form/inputs/base";

import stylesCommon from "../common/toggle.module.scss";

function ImageToggle(props: ToolToggleProps) {
    const { activeImgSnip } = useImageEditorContext();

    // this tool only works with an active image snip
    if (!activeImgSnip) {
        return null;
    }

    return <Toggle label="Resize" icon={<TbPhotoEdit />} {...props} />;
}

function ImageOverlay() {
    const { setImgRect, imgRect, setSubRect, activeImgSnip } =
        useImageEditorContext();
    const { currentOverlayWorker } = usePlacementContext();
    const [state, setState] = useState<"resize" | "cut">("resize");

    useSnipOverlay(activeImgSnip, currentOverlayWorker);

    return (
        <Overlay>
            {state === "resize" ? (
                <ResizeHandlerOverlayOutline />
            ) : (
                <CutHandlersOverlayOutline
                    onDoubleClick={() => {
                        setState((prev) =>
                            prev === "resize" ? "cut" : "resize",
                        );
                    }}
                />
            )}
            <OverlayContent>
                {state === "resize" ? (
                    <DragMove
                        className="w-100 h-100 position-absolute"
                        onDragMove={function (e: PointerEvent): void {
                            if (state === "resize") {
                                setImgRect((prev) => {
                                    const x = prev.x + e.movementX;
                                    const y = prev.y + e.movementY;
                                    return { ...prev, x, y };
                                });
                            } else {
                                setSubRect((prev) => {
                                    let x = prev.x + e.movementX;
                                    let y = prev.y + e.movementY;

                                    if (x < 0) {
                                        x = 0;
                                    }
                                    if (y < 0) {
                                        y = 0;
                                    }
                                    if (x + prev.width > imgRect.width) {
                                        x = prev.x;
                                    }
                                    if (y + prev.height > imgRect.height) {
                                        y = prev.y;
                                    }

                                    return { ...prev, x, y };
                                });
                            }
                        }}
                        onDoubleClick={() => {
                            setState((prev) =>
                                prev === "resize" ? "cut" : "resize",
                            );
                        }}
                    />
                ) : null}
            </OverlayContent>
        </Overlay>
    );
}

function ImageHelp() {
    return (
        <Help>
            <HelpHeader>
                <span>Image</span>
                <TbPhotoEdit
                    className={stylesCommon.icon}
                    style={{
                        fill: "transparent",
                    }}
                />
            </HelpHeader>
            <HelpDescription>
                Resize and cut the image. 2 x <kbd>LMB</kbd> to change between
                resizing and cutting. You may hold <kbd>Shift</kbd> to maintain
                the aspect ratio.
            </HelpDescription>
        </Help>
    );
}

function ImageSubMenu(props: ToolSubmenuProps) {
    const { resetImgSnip } = useImageEditorContext();
    return (
        <Submenu {...props}>
            <ItemSmall>
                <ResetButton onReset={resetImgSnip} />
            </ItemSmall>
        </Submenu>
    );
}

const ImageTool = {
    name: "image" as const,
    Toggle: ImageToggle,
    Help: ImageHelp,
    Overlay: ImageOverlay,
    Submenu: ImageSubMenu,
    ContextProvider: ImageEditorContextProvider,
} as ToolComponents;

export default ImageTool;
