import { useEffect, useRef, useState } from "react";

import { ArraySnip } from "@snip/snips/general/array";
import { BaseSnip } from "@snip/snips/general/base";
import { TextSnipLegacy } from "@snip/snips/general/legacy/text";
import { TextSnip } from "@snip/snips/general/text";

import { useMoveToolContext } from "../move/context";

import { useWorkerContext } from "components/context/viewer/useWorkerContext";
import { useEditorContext } from "components/editor/context";
import { OverlayWorker } from "components/editor/worker/overlayWorker";

import styles from "./interaction.module.scss";

export default function SnipToDom({
    snip,
    offset,
}: {
    snip: BaseSnip;
    offset?: [number, number];
}) {
    if (snip instanceof TextSnip) {
        return <TextSnipAsFC snip={snip} offset={offset} />;
    }
    if (snip instanceof TextSnipLegacy) {
        return <TextSnipLegacyAsFC snip={snip} offset={offset} />;
    }
    if (snip instanceof ArraySnip) {
        return <ArraySnipAsFC snip={snip} offset={offset} />;
    }

    // TODO: add overlays for other snips
    // e.g. links
    // add right click menu for snips to view some information
    return null;
}

function ArraySnipAsFC({
    snip,
    offset = [0, 0],
}: {
    snip: ArraySnip;
    offset?: [number, number];
}) {
    let legacy_offset = [0, 0];
    const childs = snip.snips.map((s, i) => {
        const o: [number, number] = [
            snip.x + legacy_offset[0] + offset[0],
            snip.y + legacy_offset[1] + offset[1],
        ];
        const c = <SnipToDom key={i} snip={s} offset={o} />;
        if (snip.legacy) {
            legacy_offset = [legacy_offset[0], legacy_offset[1] + s.height];
        }
        return c;
    });

    return <>{childs}</>;
}

/** Generate a text snip as a functional component
 * which can be used in the overlay
 *
 * @param {Snip} snip - The snip to render
 * @param {number} offsetDB - The offset of the snip in the database y coord
 * @returns {React.FC} - The functional component
 */
function TextSnipAsFC({
    snip,
    offset,
}: {
    snip: TextSnip;
    offset?: [number, number];
}) {
    const { zoom } = useMoveToolContext();
    const { overlayWorker } = useWorkerContext();
    const { pageWrappersRef } = useEditorContext();
    const [parsedText, setParsedText] = useState<string | null>(null);
    const ref = useRef<HTMLTextAreaElement>(null);

    useEffect(() => {
        const page = pageWrappersRef.current?.get(0)
            ?.children?.[0] as HTMLElement;
        const ele = ref.current;
        const z = zoom.get(0);
        if (!page || !ele || !z) return;

        // Set positions
        const worker = overlayWorker.get(0);
        const xy = _getPos(snip, worker, offset);

        ele.style.left = page.offsetLeft + xy[0] + "px";
        // No idea where the offset comes from most likely css related
        const css_offset = (snip.canvasFont.fontSize * z) / 12;
        ele.style.top = page.offsetTop + xy[1] - css_offset + "px";

        //Set styles
        ele.style.fontFamily = snip.canvasFont.fontFamily;
        ele.style.fontSize = snip.canvasFont.fontSize * z + "px";
        ele.style.lineHeight =
            snip.canvasFont.lineHeight * snip.canvasFont.fontSize * z + "px";

        ele.style.width = snip.width * z + "px";
        ele.style.height = snip.height * z + "px";

        // Parse and set text
        let text = snip.text;
        text = text
            .replaceAll(/\t/g, "        ")
            .replaceAll(/\n/g, "\\n")
            .replaceAll(/\r/g, "")
            .replaceAll(/\\n/g, String.fromCharCode(13, 10));
        console.log(text);
        setParsedText(text);
    }, [overlayWorker, pageWrappersRef, snip, zoom, offset]);

    return (
        <textarea
            ref={ref}
            className={styles.textSnip}
            value={parsedText}
            autoCorrect="off"
            wrap="off"
            readOnly
        />
    );
}

function TextSnipLegacyAsFC({
    snip,
    offset,
}: {
    snip: TextSnipLegacy;
    offset?: [number, number];
}) {
    const { zoom } = useMoveToolContext();
    const { overlayWorker } = useWorkerContext();
    const { pageWrappersRef } = useEditorContext();
    const [parsedText, setParsedText] = useState<string | null>(null);
    const ref = useRef<HTMLTextAreaElement>(null);

    useEffect(() => {
        const page = pageWrappersRef.current?.get(0)
            ?.children?.[0] as HTMLElement;
        const ele = ref.current;
        let z = zoom.get(0);
        if (!page || !ele || !z) return;
        z = z * 2.35;
        // Set positions
        const xy = _getPos(snip, overlayWorker.get(0), offset);
        ele.style.left = page.offsetLeft + xy[0] + "px";
        // No idea where the offset comes from most likely css related
        const css_offset = (snip.fontSize * z) / 6;
        ele.style.top = page.offsetTop + xy[1] - css_offset + "px";

        //Set styles
        ele.style.fontFamily = snip.fontFamily;
        ele.style.fontSize = snip.fontSize * z + "px";
        //ele.style.lineHeight = snip.lineHeight + "px";

        ele.style.width = snip.width * z + "px";
        ele.style.height = snip.height * z + "px";
        ele.style.lineHeight = snip.lineHeight * z + "px";
        // Parse and set text
        let text = snip.text;
        text = text
            .replaceAll(/\t/g, "        ")
            .replaceAll(/\n/g, "\\n")
            .replaceAll(/\r/g, "")
            .replaceAll(/\\n/g, String.fromCharCode(13, 10));
        setParsedText(text);
        console.log(text);
    }, [offset, overlayWorker, pageWrappersRef, snip, zoom]);

    return (
        <textarea
            ref={ref}
            className={styles.textSnip}
            value={parsedText}
            autoCorrect="off"
            wrap="off"
            readOnly
        />
    );
}

function _getPos(
    snip: BaseSnip,
    worker: OverlayWorker,
    offset?: [number, number],
): [number, number] {
    const xy = worker.dbCoords_2_canvasCoords([snip.x, snip.y]);

    if (offset) {
        const o = worker.dbCoords_2_canvasCoords(offset);
        xy[0] += o[0];
        xy[1] += o[1];
    }

    return xy as [number, number];
}
