import { useCallback, useEffect, useRef } from "react";
import { createPortal } from "react-dom";
import { TbPoint } from "react-icons/tb";

import { Toggle } from "../common/toggle";
import { ToolComponents, ToolToggleProps } from "../types";
import { PointerContextProvider, usePointerContext } from "./context";

import { useWorkerContext } from "components/context/viewer/useWorkerContext";
import { useEditorContext } from "components/editor/context";

function PointerToggle(props: ToolToggleProps) {
    return <Toggle label="Pointer" icon={<TbPoint />} {...props} />;
}

function PointerHelp() {
    return null;
}

function PointerOverlay() {
    const { pageWrappersRef } = useEditorContext();
    const cursorRef = useRef<HTMLDivElement>();
    const { overlayWorker } = useWorkerContext();
    /** The move event gets propagated threw the socket connection
     * to other clients which joined the current page. Positions
     * are send in db coords.
     */
    const handleMove = useCallback(
        (e: PointerEvent, idx: number) => {
            // Transmit position to other clients in real-time
            const xy: [number, number] = [e.offsetX, e.offsetY];
            const w = overlayWorker.get(idx);
            if (!w) return;
            const xy_db = w.canvasCoords_2_dbCoords(xy);
            w.laserPointerHandler.sendPos(xy_db);
        },
        [overlayWorker],
    );

    useEffect(() => {
        const pages = Array.from(
            pageWrappersRef.current?.values(),
            (wrapper) => wrapper.children[0] as HTMLDivElement,
        );

        const moveHandlers = pages.map((page, idx) => {
            return (e) => handleMove(e, idx);
        });

        pages.forEach((page, idx) => {
            page.addEventListener("pointermove", moveHandlers[idx]);

            page.style.cursor = "crosshair";
        });

        return () => {
            pages.forEach((page, idx) => {
                page.removeEventListener("pointermove", moveHandlers[idx]);

                page.style.cursor = "default";
            });
        };
    }, [handleMove, pageWrappersRef]);

    return null;
}

const PointerTool: ToolComponents = {
    name: "laserPointer" as const,
    Toggle: PointerToggle,
    Help: PointerHelp,
    Overlay: PointerOverlay,
    ContextProvider: PointerContextProvider,
};

export default PointerTool;
