import { useDebounce } from "lib/hooks/useDebounce";
import {
    createContext,
    useCallback,
    useContext,
    useEffect,
    useState,
} from "react";

import { ToolContextProps } from "../types";

import { useWorkerContext } from "components/context/viewer/useWorkerContext";
import { useEditorContext } from "components/editor/context";

type page_idx = number;

interface MoveContextI {
    _zoom: Map<page_idx, number>;
    zoom: Map<page_idx, number>;
    setZoom: (page: page_idx, zoom: (prev: number) => number) => void;

    // The scroll position of the page
    scrollPosition: Map<page_idx, [number, number]>;
    setScrollPosition: (page: page_idx, pos: [number, number]) => void;
    setScrollOffset: (page: page_idx, offset: [number, number]) => void;

    // Helpers
    centerPage: (page_idx: page_idx, x?: boolean, y?: boolean) => void;
    autoFitPage: (page_idx: page_idx, width: boolean, height: boolean) => void;
}

const MoveToolContext = createContext<MoveContextI>({
    _zoom: new Map(),
    zoom: new Map(),
    setZoom: () => {},
    scrollPosition: new Map(),
    setScrollPosition: () => {},
    setScrollOffset: () => {},
    centerPage: () => {},
    autoFitPage: () => {},
});

export function MoveToolContextProvider({ children }: ToolContextProps) {
    const { pageWrappersRef } = useEditorContext();
    const { worker, overlayWorker } = useWorkerContext();

    /* -------------------------------------------------------------------------- */
    /*                           evertyhing zoom related                          */
    /* -------------------------------------------------------------------------- */

    const [zoom, _setZoom] = useState<Map<page_idx, number>>(new Map([[0, 1]]));
    const setZoom = useCallback(
        (page: page_idx, zoom: (prev: number) => number) => {
            _setZoom((prev) => {
                const next = new Map(prev);
                const prev_val = prev.get(page) || 1;
                let next_val = zoom(prev_val);
                if (next_val <= 0.25) next_val = 0.25;
                if (next_val >= 4) next_val = 4;
                next.set(page, next_val);

                // apply the zoom to the pages
                // we have to use the width and height as we want
                // to dynamically redraw the canvas
                // scale also works but it does not redraw the canvas
                // In the correct width and heigh
                const pageDom = pageWrappersRef.current.get(page);
                if (!pageDom) return next;

                // The overlay is the wrapper around the
                // canvas elements
                const overlay = pageDom.children[0] as HTMLDivElement;
                if (!overlay) return;

                const newWidth = 1400 * next_val;
                const newHeight = 2000 * next_val;
                overlay.style.width = `${newWidth}px`;
                overlay.style.height = `${newHeight}px`;
                // Always center the overlay
                overlay.style.top = `calc(50% - ${newHeight / 2}px)`;
                overlay.style.left = `calc(50% - ${newWidth / 2}px)`;
                worker.get(page)?.resize(newWidth, newHeight);
                overlayWorker.get(page)?.resize(newWidth, newHeight);

                return next;
            });
        },
        [overlayWorker, pageWrappersRef, worker],
    );

    const debouncedZoom = useDebounce(zoom, 500);

    /* -------------------------------------------------------------------------- */
    /*        everything scroll / position related i.e. allows to move page       */
    /* -------------------------------------------------------------------------- */

    const [scrollPosition, _setScrollPosition] = useState<
        Map<page_idx, [number, number]>
    >(new Map());

    const setScrollPosition = useCallback(
        (page: page_idx, pos: [number, number]) => {
            _setScrollPosition((prev) => {
                const next = new Map(prev);
                next.set(page, pos);
                return next;
            });
        },
        [],
    );
    const setScrollOffset = useCallback(
        (page: page_idx, offset: [number, number]) => {
            _setScrollPosition((prev) => {
                const next = new Map(prev);
                let pos = next.get(page);

                // Workaround for initialized with other
                // scroll pos
                if (!pos) {
                    const p = pageWrappersRef.current.get(page);
                    pos = p ? [p.scrollLeft, p.scrollTop] : [0, 0];
                }
                next.set(page, [pos[0] + offset[0], pos[1] + offset[1]]);
                return next;
            });
        },
        [pageWrappersRef],
    );

    useEffect(() => {
        // apply the scroll position to the pages
        scrollPosition.forEach((value, key) => {
            const page = pageWrappersRef.current.get(key);
            if (page) {
                const parent = page.parentElement;
                parent.scrollTo({
                    left: value[0],
                    top: value[1],
                    behavior: "instant",
                });
            }
        });
    }, [pageWrappersRef, scrollPosition]);

    /* -------------------------------------------------------------------------- */
    /*                                some helpers                                */
    /* -------------------------------------------------------------------------- */

    /**
     * Centers the specified page within its parent view.
     * @param page_idx - The index of the page to center.
     * @param x - Whether to center the page horizontally.
     * @param y - Whether to center the page vertically.
     *
     * @returns void
     */
    const centerPage = useCallback(
        (page_idx: page_idx, x: boolean = true, y: boolean = true) => {
            const page = pageWrappersRef.current.get(page_idx);
            if (!page) return;
            const view = page.parentElement;
            if (page) {
                _setScrollPosition((prev) => {
                    const next = new Map(prev);
                    next.set(page_idx, [
                        x
                            ? page.clientWidth / 2 - view.clientWidth / 2
                            : prev.get(page_idx)?.[0],
                        y
                            ? page.clientHeight / 2 - view.clientHeight / 2
                            : prev.get(page_idx)?.[1],
                    ]);
                    return next;
                });
            }
        },
        [pageWrappersRef],
    );

    const autoFitPage = useCallback(
        (page_idx: page_idx, width: boolean = true, height: boolean = true) => {
            const page = pageWrappersRef.current.get(page_idx);
            if (!page) return;
            const view = page.parentElement;
            if (page) {
                let zoom = 1;
                if (width && height) {
                    zoom = Math.min(
                        view.clientWidth / 1400,
                        view.clientHeight / 2000,
                    );
                } else if (width) {
                    zoom = view.clientWidth / 1400;
                } else if (height) {
                    zoom = view.clientHeight / 2000;
                }
                setZoom(page_idx, () => zoom);

                //Compute fit position
                const newWidth = 1400 * zoom;
                const newHeight = 2000 * zoom;

                // compute center positions
                const overhang_y = newHeight - view.clientHeight;
                const overhang_x = newWidth - view.clientWidth;
                let x_pos = (page.clientWidth - view.clientWidth) / 2;
                let y_pos = (page.clientHeight - view.clientHeight) / 2;

                // depending on parameter subtract the overhang
                if (height && overhang_x > 0) {
                    x_pos = x_pos - (newWidth - view.clientWidth) / 2;
                }
                if (width && overhang_y > 0) {
                    y_pos = y_pos - (newHeight - view.clientHeight) / 2;
                }

                setScrollPosition(page_idx, [x_pos, y_pos]);
                //overlayWorker.get(page_idx)?.resize(newWidth, newHeight);
            }
        },
        [pageWrappersRef, setZoom, setScrollPosition],
    );

    const scrollPositionDebounced = useDebounce(scrollPosition, 500);

    return (
        <MoveToolContext.Provider
            value={{
                _zoom: zoom,
                zoom: debouncedZoom,
                setZoom,
                scrollPosition: scrollPositionDebounced,
                setScrollPosition,
                setScrollOffset,
                centerPage,
                autoFitPage,
            }}
        >
            {children}
        </MoveToolContext.Provider>
    );
}

export function useMoveToolContext() {
    return useContext(MoveToolContext);
}
