import { useCallback, useEffect, useRef } from "react";

import { useDrag } from "../common/drag";
import { useMoveToolContext } from "./context";

import { useEditorContext } from "components/editor/context";

type PointerTypes = "mouse" | "touch" | "pen";

/**
 * This handler is responsible for moving the pages
 *
 * @returns
 */
export function MoveHandler({
    pointerTypes,
}: {
    pointerTypes?: PointerTypes[];
}) {
    const { pageWrappersRef } = useEditorContext();
    const { setScrollOffset, centerPage, autoFitPage } = useMoveToolContext();

    const currentDraggingPage = useRef<number>(null);
    const prevPos = useRef<[number, number]>(null);

    const onDragMove = useCallback(
        (e: PointerEvent) => {
            setScrollOffset(currentDraggingPage.current, [
                prevPos.current[0] - e.clientX,
                prevPos.current[1] - e.clientY,
            ]);

            prevPos.current = [e.clientX, e.clientY];
        },
        [setScrollOffset],
    );

    const prevCursor = useRef<string>("");
    const onPointerDown = useCallback(() => {
        pageWrappersRef.current.forEach((value) => {
            // cursor styling
            prevCursor.current = value.style.cursor;
            value.style.cursor = "grabbing";
            value.style.userSelect = "none";
        });
    }, [pageWrappersRef]);

    const onPointerUp = useCallback(() => {
        pageWrappersRef.current.forEach((value) => {
            value.style.cursor = prevCursor.current;
            console.log("remove", prevCursor.current);
            value.style.removeProperty("user-select");
        });
    }, [pageWrappersRef]);

    const [handlePointerDown] = useDrag({
        pointerTypes,
        onDragMove,
        onPointerDown,
        onPointerUp,
    });

    useEffect(() => {
        console.log("MoveHandler mounted");
        return () => {
            console.log("MoveHandler unmounted");
        };
    }, [autoFitPage, centerPage, handlePointerDown]);

    useEffect(() => {
        const pageElements = [];
        pageWrappersRef.current.forEach((value) => {
            pageElements.push(value);
        });

        const handlers = [];
        pageElements.forEach((value, idx) => {
            handlers.push((e: PointerEvent) => {
                prevPos.current = [e.clientX, e.clientY];
                currentDraggingPage.current = idx;

                handlePointerDown(e);
            });

            // Styles
            if (
                !pointerTypes ||
                (pointerTypes && pointerTypes.includes("mouse"))
            ) {
                value.style.cursor = "grab";
                value.style.userSelect = "none";
            }
        });

        handlers.forEach((handler, idx) => {
            pageElements[idx].addEventListener("pointerdown", handler);
        });

        //Register double click to fit page
        const handlers_db = pageElements.map((value, idx) => {
            return (e: MouseEvent) => {
                e.preventDefault();
                e.stopPropagation();
                centerPage(idx, true, true);
            };
        });

        const handlers_aux = pageElements.map((value, idx) => {
            return (e: MouseEvent) => {
                e.preventDefault();
                e.stopPropagation();
                autoFitPage(idx, true, true);
            };
        });

        pageElements.forEach((value, idx) => {
            value.addEventListener("dblclick", handlers_db[idx]);
            value.addEventListener("auxclick", handlers_aux[idx]);
        });

        return () => {
            handlers.forEach((handler, idx) => {
                pageElements[idx].removeEventListener("pointerdown", handler);
                pageElements[idx].style.cursor = "default";
                pageElements[idx].style.removeProperty("user-select");
            });

            pageElements.forEach((value, idx) => {
                value.removeEventListener("dblclick", handlers_db[idx]);
                value.removeEventListener("auxclick", handlers_aux[idx]);
            });
        };
    }, [
        autoFitPage,
        centerPage,
        handlePointerDown,
        pageWrappersRef,
        pointerTypes,
    ]);

    return null;
}
