import useUser from "lib/hooks/useUser";
import {
    createContext,
    Dispatch,
    SetStateAction,
    useCallback,
    useContext,
    useEffect,
    useMemo,
    useRef,
    useState,
} from "react";

import { AllowedFonts } from "@snip/database/fonts/internal";
import { isBrowser } from "@snip/utils/environment";

import { useMoveToolContext } from "../move/context";
import { ToolContextProps } from "../types";

import { useEditorContext } from "components/editor/context";

export interface TextStyles {
    /** The font family of the text.
     * this is extracted via the sql_to_ts script on
     * startup from the database. I.e. if a new font is added
     * to the database, the server needs to be restarted for it
     * to be available.
     */
    fontFamily: AllowedFonts;

    /** The font size of the text.
     * This is in db units.
     */
    fontSize: number;

    /** The line height of the text.
     * This is in db units.
     */
    lineHeight: number;

    /** The line wrap of the text. I.e. how many points
     * are allowed in one line before the text is wrapped.
     * This is in db units.
     *
     * This is parsed using the width state!
     */
    lineWrap: number;
}

export interface TextToolContextI {
    /** The text styles define the current text styles
     * that are applied to the text if it is created.
     * This is in db units.
     */
    textStyles: TextStyles;
    setTextStyles: (styles: Partial<TextStyles>) => void;
    /** Transformed TextStyles to dom styles
     */
    domTextStyles: Record<string, string>;
    setDomTextStyles: (styles: Partial<TextStyles>) => void;

    /** The currently active text snip. Useful in some
     * cases, e.g. when the user wants to edit the text
     * snip before it is finally placed.
     */
    text: string;
    setText: Dispatch<SetStateAction<string>>;

    /** The page on which the text should be placed */
    pageIdx: number;
    setPageIdx: Dispatch<SetStateAction<number>>;

    incrementWrapByDrag: (x_mov: number) => void;
}

const TextToolContext = createContext<TextToolContextI>({
    textStyles: {
        fontFamily: "Arial",
        fontSize: 12,
        lineHeight: 1.25,
        lineWrap: 400,
    },
    setTextStyles: () => {},
    domTextStyles: {},
    setDomTextStyles: () => {},
    text: "",
    setText: () => {},
    pageIdx: 0,
    setPageIdx: () => {},
    incrementWrapByDrag: () => {},
});

export function useTextToolContext() {
    return useContext(TextToolContext) as TextToolContextI;
}

export function TextToolContextProvider({ children }: ToolContextProps) {
    const { setOverlayRect } = useEditorContext();
    const { zoom } = useMoveToolContext();
    const [text, setText] = useState<string>("");

    /**
     *  Styles in db units
     */
    const [textStyles, _setTextStyles] = useState<TextStyles>({
        fontFamily: "Arial",
        fontSize: 12,
        lineHeight: 1.25,
        lineWrap: 400,
    });

    const setTextStyles = useCallback(
        (styles: Partial<TextStyles>) => {
            if (styles.lineWrap !== undefined) {
                setOverlayRect((prev) => {
                    console.log("setting overlay rect", prev);
                    return {
                        ...prev,
                        width: styles.lineWrap,
                    };
                });
            }
            _setTextStyles((prev) => {
                return {
                    ...prev,
                    ...styles,
                };
            });
        },
        [setOverlayRect],
    );

    /** Dom text transfered
     * from db units
     */
    const PAGE_IDX = 0;
    const scale = zoom.get(PAGE_IDX) || 1;
    const domTextStyles = useMemo(() => {
        //compute min width given the computed width
        const z = scale * 2.35;
        const textWidth = getTextWidth(
            "AB",
            textStyles.fontFamily,
            textStyles.fontSize * z,
        );
        return {
            fontFamily: textStyles.fontFamily,
            fontSize: textStyles.fontSize * z + "px",
            lineHeight: textStyles.lineHeight * textStyles.fontSize * z + "px",
            width: "100% !important",
            minHeight:
                textStyles.fontSize *
                    textStyles.lineHeight *
                    z *
                    text.split("\n").length -
                1 +
                "px",
            minWidth: textWidth + "px",
        };
    }, [
        textStyles.fontFamily,
        textStyles.fontSize,
        textStyles.lineHeight,
        scale,
        text,
    ]);

    const setDomTextStyles = useCallback(
        (dom_styles: Partial<TextStyles>) => {
            const z = scale * 2.35;
            const styles: Partial<TextStyles> = {};
            if (dom_styles.fontSize !== undefined) {
                styles.fontSize = dom_styles.fontSize / z;
            }
            if (dom_styles.lineHeight !== undefined) {
                styles.lineHeight = dom_styles.lineHeight / z;
            }
            if (dom_styles.lineWrap !== undefined) {
                styles.lineWrap = dom_styles.lineWrap / scale;
            }
            if (dom_styles.fontFamily !== undefined) {
                styles.fontFamily = dom_styles.fontFamily as AllowedFonts;
            }

            setTextStyles(styles);
        },
        [setTextStyles, scale],
    );

    const incrementWrapByDrag = useCallback(
        (x_mov: number) => {
            setOverlayRect((prev) => {
                return {
                    ...prev,
                    width: prev.width + x_mov,
                };
            });

            _setTextStyles((prev) => {
                return {
                    ...prev,
                    lineWrap: prev.lineWrap + x_mov / scale,
                };
            });
        },
        [setOverlayRect, scale],
    );

    // Which page to place the text
    // TODO: Add way to switch pages
    const [pageIdx, setPageIdx] = useState<number>(0);
    const { config } = useUser(true);
    const mounted = useRef(false);

    //Initial sync from db
    useEffect(() => {
        if (mounted.current || !config) return;
        mounted.current = true;

        setTextStyles({
            fontFamily: config.text_font as AllowedFonts,
            fontSize: config.text_fontSize,
            lineHeight: config.text_lineHeight,
            lineWrap: config.text_lineWrap,
        });

        return () => {
            mounted.current = false;
        };
    }, [config, setTextStyles]);

    return (
        <TextToolContext.Provider
            value={{
                textStyles,
                setTextStyles,
                domTextStyles,
                setDomTextStyles,
                text,
                setText,
                pageIdx,
                setPageIdx,
                incrementWrapByDrag,
            }}
        >
            {children}
        </TextToolContext.Provider>
    );
}

interface TextWidthI {
    (text: string, font: string, size: number): number;
    canvas?: HTMLCanvasElement;
}
const getTextWidth: TextWidthI = function (text, font, size) {
    // re-use canvas object for better performance
    if (!isBrowser()) {
        return 0;
    }
    const canvas =
        getTextWidth.canvas ||
        (getTextWidth.canvas = document.createElement("canvas"));
    const context = canvas.getContext("2d");
    context.font = size + "px " + font;
    const metrics = context.measureText(text);
    return metrics.width;
};
