import { useEffect, useRef } from "react";
import { BsFonts } from "react-icons/bs";

import { AllowedFonts, FONT_URLS } from "@snip/database/fonts/internal";
import { TextSnip } from "@snip/snips/general/text";

import DragMove from "../common/drag";
import {
    Help,
    HelpDescription,
    HelpHeader,
    HelpShortcuts,
} from "../common/help";
import {
    Overlay,
    OverlayContent,
    OverlayMover,
    OverlayOutline,
    OverlayPlaceButton,
} from "../common/overlay";
import { ItemSmall, Submenu } from "../common/submenu";
import { Toggle } from "../common/toggle";
import { useToolbarContext } from "../toolbar";
import { ToolComponents, ToolSubmenuProps, ToolToggleProps } from "../types";
import { TextToolContextProvider, useTextToolContext } from "./context";
import { KeyHandler } from "./keyHandler";

import { usePagesContext } from "components/context/socket/usePagesContext";
import { useWorkerContext } from "components/context/viewer/useWorkerContext";
import { useEditorContext } from "components/editor/context";
import { PageWorker } from "components/editor/worker/pageWorker";
import { NumberInput } from "components/form/inputs/number";
import { DropdownInput } from "components/form/inputs/option";

import styles from "./text.module.scss";

function TextToggle(props: ToolToggleProps) {
    return (
        <Toggle
            label="Move"
            icon={
                <BsFonts
                    style={{
                        stroke: "none",
                    }}
                />
            }
            {...props}
        />
    );
}

function TextSubMenu(props: ToolSubmenuProps) {
    const { textStyles, setTextStyles } = useTextToolContext();

    useEffect(() => {
        const inp = document.getElementById(
            "font_size_text_input",
        ) as HTMLInputElement;
        if (inp) {
            inp.value = textStyles.fontSize.toString();
        }

        const inp2 = document.getElementById(
            "line_wrap_text_input",
        ) as HTMLInputElement;
        if (inp2) {
            inp2.value = textStyles.lineWrap.toFixed();
        }

        const inp3 = document.getElementById(
            "line_height_text_input",
        ) as HTMLInputElement;
        if (inp3) {
            inp3.value = textStyles.lineHeight.toString();
        }

        return () => {};
    }, [textStyles]);

    return (
        <Submenu {...props} className={styles.menu}>
            <ItemSmall className={styles.font_family} style={{ zIndex: 3 }}>
                <DropdownInput
                    label="Font Family"
                    value={textStyles.fontFamily}
                    options={FONT_URLS.map(([, name]) => ({
                        eventKey: name,
                        label: (
                            <span
                                style={{
                                    fontFamily: name,
                                }}
                            >
                                {name}
                            </span>
                        ),
                    }))}
                    onSelect={(name) => {
                        setTextStyles({
                            fontFamily: name as AllowedFonts,
                        });
                    }}
                />
            </ItemSmall>
            <ItemSmall className={styles.font_size}>
                <NumberInput
                    id="font_size_text_input"
                    label="Font Size"
                    min={6}
                    max={72}
                    step={1}
                    defaultValue={textStyles.fontSize}
                    onInput={(e) => {
                        setTextStyles({
                            fontSize: parseInt(e.target.value),
                        });
                    }}
                    onReset={() => {
                        setTextStyles({
                            fontSize: 12,
                        });
                    }}
                />
            </ItemSmall>
            <ItemSmall className={styles.line_wrap}>
                <NumberInput
                    id="line_wrap_text_input"
                    label="Line Wrap"
                    min={50}
                    max={1000}
                    step={1}
                    defaultValue={textStyles.lineWrap.toFixed()}
                    onInput={(e) => {
                        setTextStyles({
                            lineWrap: parseInt(e.target.value),
                        });
                    }}
                    onReset={() => {
                        setTextStyles({
                            lineWrap: 400,
                        });
                    }}
                />
            </ItemSmall>
            <ItemSmall className={styles.line_height}>
                <NumberInput
                    id="line_height_text_input"
                    label="Line Height"
                    min={0.5}
                    max={3}
                    step={0.1}
                    defaultValue={textStyles.lineHeight}
                    onInput={(e) => {
                        setTextStyles({
                            lineHeight: parseFloat(e.target.value),
                        });
                    }}
                    onReset={() => {
                        setTextStyles({
                            lineHeight: 1.25,
                        });
                    }}
                />
            </ItemSmall>
        </Submenu>
    );
}

function TextOverlay() {
    const areaRef = useRef<HTMLTextAreaElement>(null);
    const {
        text,
        setText,
        domTextStyles,

        incrementWrapByDrag,
    } = useTextToolContext();
    const { setOverlayRect } = useEditorContext();

    useEffect(() => {
        if (areaRef.current) {
            textAreaAdjust(areaRef.current);
        }
    }, [text]);

    return (
        <>
            <Overlay>
                <OverlayOutline>
                    <OverlayMover
                        onMove={(mov) => {
                            setOverlayRect((prev) => ({
                                ...prev,
                                x: prev.x + mov[0],
                                y: prev.y + mov[1],
                            }));
                        }}
                    />
                    <PlaceTextButton />
                    <DragMove
                        className={styles.resizerWrap}
                        onDragMove={(e) => {
                            console.log(e.movementX);
                            incrementWrapByDrag(e.movementX);
                        }}
                    />
                </OverlayOutline>
                <OverlayContent>
                    <textarea
                        ref={areaRef}
                        className={styles.textarea}
                        value={text}
                        //Scale text styles to the overlay
                        style={domTextStyles}
                        onInput={(
                            e: React.ChangeEvent<HTMLTextAreaElement>,
                        ) => {
                            setText(e.target.value);
                            textAreaAdjust(e.target);
                        }}
                    />
                </OverlayContent>
            </Overlay>
            <KeyHandler />
        </>
    );
}

function PlaceTextButton() {
    const { worker, overlayWorker } = useWorkerContext();
    const { pages } = usePagesContext();
    const { pageWrappersRef, activePageIds, overlayRect } = useEditorContext();
    const { textStyles, text, pageIdx, setText } = useTextToolContext();
    const { setActiveTool } = useToolbarContext();
    // TODO: somehow link with normal place tool
    // it is a bit inconsistent
    return (
        <OverlayPlaceButton
            onPlace={() => {
                // Only place if text is not empty (only contains spaces)
                if (text === "" || text.trim() === "") return;
                const page = pages.get(activePageIds[pageIdx]);
                const w = worker.get(pageIdx) as PageWorker;
                const o = overlayWorker.get(pageIdx);
                if (!page || !w || !o) return;

                // Transform x and y to db coords
                const wrapper = pageWrappersRef.current.get(pageIdx);
                const pageDiv = wrapper.children[0] as HTMLDivElement;
                if (!pageDiv) return;
                // Translate x and y to the page
                const xN = overlayRect.x - pageDiv.offsetLeft;
                const yN = overlayRect.y - pageDiv.offsetTop;
                const pos = o.canvasCoords_2_dbCoords([xN, yN]);

                // create snip data
                const textSnip = new TextSnip({
                    ...textStyles,
                    text,
                    book_id: page.book_id,
                    x: Math.round(pos[0]),
                    // No idea where the offset comes from most likely css related
                    y: Math.round(pos[1]) + textStyles.fontSize / 6,
                });
                console.log(textSnip.to_data());
                w.placeOnPage(textSnip.to_data());
                setText("");
                setActiveTool("interaction");
            }}
        />
    );
}

function textAreaAdjust(element) {
    element.style.height = "auto";
    element.style.height = element.scrollHeight + "px";
}

function TextHelp() {
    return (
        <Help>
            <HelpHeader>
                <span>Text</span>
                <BsFonts style={{ stroke: "none" }} />
            </HelpHeader>
            <HelpDescription>
                The text tool allows you to add text to the page. It is
                relatively simple, but allows for some basic formatting options.
                Placement has to be confirmed!
            </HelpDescription>
            <HelpShortcuts
                shortcuts={[
                    ["t", "Toggle tool"],
                    ["o", "Font size down"],
                    ["i", "Font size up"],
                ]}
            />
        </Help>
    );
}

const TextTool: ToolComponents = {
    name: "text" as const,
    Toggle: TextToggle,
    Submenu: TextSubMenu,
    Help: TextHelp,
    Overlay: TextOverlay,
    ContextProvider: TextToolContextProvider,
};

export default TextTool;
