"use client";
import {
    createContext,
    Dispatch,
    ReactNode,
    SetStateAction,
    useContext,
    useEffect,
    useMemo,
    useState,
} from "react";
import { Button } from "react-bootstrap";
import { createPortal } from "react-dom";
import { BsFonts } from "react-icons/bs";
import { TbArrowsMove, TbHelpSmall, TbPointer } from "react-icons/tb";

import { useEditorContext } from "../context";
import {
    Help as HelpWrapper,
    HelpDescription,
    HelpHeader,
} from "./common/help";
import { ToolsContextProvider } from "./context";
import { PenEmpty } from "./doodle/svgs";
import { getToolComponents } from "./toolsFactory";
import { Tool, ToolComponents } from "./types";

import styles from "./toolbar.module.scss";

export function Toolbar({
    tools = ["interaction", "movePage", "doodle", "text", "placement", "image"],
}: {
    tools?: Tool[];
}) {
    const { overlayContainer } = useEditorContext();

    const toolsComponents = useMemo(() => {
        return tools.map((tool) => getToolComponents(tool));
    }, [tools]);

    return (
        <ToolbarContextProvider>
            <ToolsContextProvider
                contextProviders={toolsComponents.map((t) => t.ContextProvider)}
            >
                <div className={styles.toolbar} data-position="left">
                    <Menu>{toolsComponents}</Menu>
                    <Overlay overlayContainer={overlayContainer}>
                        {toolsComponents}
                    </Overlay>
                    <Help>{toolsComponents}</Help>
                </div>
            </ToolsContextProvider>
        </ToolbarContextProvider>
    );
}

/* -------------------------------------------------------------------------- */
/*                         The context for the toolbar                        */
/* -------------------------------------------------------------------------- */
// Tools might use this

interface ToolbarContext {
    activeTool: Tool;
    setActiveTool: Dispatch<SetStateAction<Tool>>;
    activeSubmenu: Tool;
    setActiveSubmenu: Dispatch<SetStateAction<Tool>>;
    showHelp: boolean;
    setShowHelp: Dispatch<SetStateAction<boolean>>;
}

const ToolbarContext = createContext<ToolbarContext>({
    activeTool: "interaction",
    setActiveTool: () => {},
    activeSubmenu: null,
    setActiveSubmenu: () => {},
    showHelp: false,
    setShowHelp: () => {},
});

function ToolbarContextProvider({ children }: { children: ReactNode }) {
    const [showHelp, setShowHelp] = useState(false);
    const [activeTool, setActiveTool] = useState<Tool>("interaction");
    const [activeSubmenu, setActiveSubmenu] = useState<Tool>(null);

    const { activeSnipId } = useEditorContext();

    /** On queued snip click set specific tools
     * to be active or inactive
     */
    useEffect(() => {
        if (activeSnipId) {
            setActiveTool("placement");
        } else {
            setActiveTool("interaction");
        }
    }, [activeSnipId]);

    return (
        <ToolbarContext.Provider
            value={{
                showHelp,
                setShowHelp,
                activeTool,
                setActiveTool,
                activeSubmenu,
                setActiveSubmenu,
            }}
        >
            {children}
        </ToolbarContext.Provider>
    );
}

export function useToolbarContext() {
    return useContext(ToolbarContext);
}

/* -------------------------------------------------------------------------- */
/*                   Helpers to wrap the tools and the help                   */
/* -------------------------------------------------------------------------- */

/**
 * Renders a menu component with toggleable tools and submenus.
 *
 * @param children - An array of tool components.
 * @returns The rendered menu component.
 */
function Menu({ children }: { children: ToolComponents[] }) {
    const { activeTool, setActiveTool, activeSubmenu, setActiveSubmenu } =
        useToolbarContext();

    function toggleTool(tool) {
        setActiveTool((prev) => (prev === tool ? null : tool));
    }

    function toggleSubmenu(submenu) {
        setActiveSubmenu((prev) => (prev === submenu ? null : submenu));
    }

    return (
        <div className={styles.menu}>
            <div className={styles.tools}>
                {children.map((tool) => {
                    return (
                        <tool.Toggle
                            key={tool.name}
                            isActive={activeTool === tool.name}
                            setActive={() => toggleTool(tool.name)}
                            isExpanded={activeSubmenu === tool.name}
                            setExpanded={() => toggleSubmenu(tool.name)}
                        />
                    );
                })}
            </div>
            <div className={styles.submenu} data-show={activeSubmenu !== null}>
                {children.map((tool) => {
                    if (!tool.Submenu) {
                        return null;
                    }
                    return (
                        <tool.Submenu
                            key={tool.name}
                            isExpanded={activeSubmenu === tool.name}
                            setExpanded={() => toggleSubmenu(tool.name)}
                        />
                    );
                })}
            </div>
        </div>
    );
}

/**
 * Renders the overlay component based on the active tool.
 *
 * @param children - An array of ToolComponents.
 * @param overlayContainer - Optional. The container element for the overlay.
 * @param overlayRect - The rectangle representing the overlay.
 * @param setOverlayRect - A function to update the overlay rectangle.
 * @returns The rendered overlay component or null if no overlay is found.
 */
function Overlay({
    children,
    overlayContainer,
}: {
    children: ToolComponents[];
    overlayContainer?: HTMLElement;
}) {
    const { activeTool } = useToolbarContext();
    const [mounted, setMounted] = useState(false);

    useEffect(() => setMounted(true), []);

    // Find the overlay component
    const Overlay = children.find((c) => c.name === activeTool)?.Overlay;

    return mounted && Overlay
        ? createPortal(<Overlay />, overlayContainer || document.body)
        : null;
}

function Help({ children }: { children: ToolComponents[] }) {
    const { showHelp, setShowHelp, activeTool } = useToolbarContext();

    // Find the help component
    const Help =
        children.find((c) => c.name === activeTool)?.Help || HelpToolbar;

    return (
        <div className={styles.help} data-expanded={showHelp}>
            <div className={styles.content}>
                <Help />
            </div>
            <Button
                variant="none"
                className={styles.expandBtn}
                onClick={() => setShowHelp(!showHelp)}
            >
                {showHelp ? <span>Hide</span> : <TbHelpSmall size={28} />}
            </Button>
        </div>
    );
}

function HelpToolbar() {
    return (
        <HelpWrapper>
            <HelpHeader>
                <span>Toolbar</span>
                <div className={styles.defHelpIcon}>
                    <TbPointer style={{ stroke: "var(--bs-secondary" }} />
                    <TbArrowsMove style={{ stroke: "var(--bs-secondary" }} />
                    <PenEmpty style={{ fill: "var(--bs-secondary" }} />
                    <BsFonts
                        style={{ fill: "var(--bs-secondary", stroke: "none" }}
                    />
                </div>
            </HelpHeader>
            <HelpDescription>
                The toolbar allows you to select different tools to interact
                with the page. Tools may have a submenu which can be opened by
                using the small arrow or by <kbd>RMB</kbd> click.
            </HelpDescription>
        </HelpWrapper>
    );
}
