"use client";
import { CanvasHTMLAttributes, useEffect, useRef } from "react";

import { debounce } from "@snip/utils";

import { RenderWorker } from "../worker/abc";

/** Creates a canvas element bound to a worker
 * using offscreen canvas. This also
 * adds basic functionality like resizing
 * and visibility change
 *
 * @param workerRef - The workers reference
 * @param idx - The index of the worker/page
 */
export function Canvas({
    worker,
    ...props
}: {
    worker: RenderWorker;
} & CanvasHTMLAttributes<HTMLCanvasElement>) {
    const canvasRef = useRef<HTMLCanvasElement>(null);
    const mounted = useRef(false);

    /**
     * On mount we create a new offscreen canvas
     * and transfer it to the worker
     */
    useEffect(() => {
        if (!worker || !canvasRef.current || mounted.current) return;
        // Setup the canvas
        const canvas = canvasRef.current;
        mounted.current = true;

        worker.setupCanvas(canvas);

        const debounce_resize = debounce((width, height) => {
            worker.resize(width, height);
        }, 200);

        //Attach resize observer
        const observer = new ResizeObserver((entries) => {
            const { width, height } = entries[0].contentRect;
            console.log("observe", width, height);
            debounce_resize(
                Math.round(width * 100) / 100,
                Math.round(height * 100) / 100,
            );
        });
        observer.observe(canvas);

        //Attach visibility change
        function visibilityChange() {
            if (document.visibilityState === "visible") {
                worker.rerender();
            }
        }
        document.addEventListener("visibilitychange", visibilityChange);

        // Cleanup
        return () => {
            mounted.current = false;
            observer.disconnect();
            document.removeEventListener("visibilitychange", visibilityChange);
        };
    }, [mounted, worker]);

    return <canvas ref={canvasRef} {...props} />;
}
