"use client";
import { useEffect } from "react";

import { PageData } from "@snip/database/types";

import { useEditorContext } from "../context";
import { Canvas } from "./canvas";

import { useWorker } from "components/context/viewer/useWorkerContext";

import styles from "./viewer.module.scss";

type PageProps = {
    pageData: PageData;
    idx?: number;
} & React.HTMLAttributes<HTMLDivElement>;

export function Page({ idx = 0, pageData, ...props }: PageProps) {
    /** We register each page to the editor if
     * it exists.
     */
    const {
        pageWrappersRef,
        setPageWrapperRef,
        overlayOnPageIdx,
        setOverlayContainer,
        setOverlayRect,
    } = useEditorContext();

    /** Each page needs
     * a worker to render it
     * and an overlay worker
     * to render the overlay
     *
     * To support multiple pages
     * we use an index to get the
     * correct worker(s)
     */
    const { worker, overlayWorker } = useWorker(idx);

    /**
     * Setup or change page
     * the setup function checks
     * if the page has changed and changes
     * the worker accordingly
     */
    useEffect(() => {
        if (!worker || !pageData) return;
        worker.setupPage(pageData);
        overlayWorker.setupPage(pageData);
    }, [worker, pageData, overlayWorker, idx]);

    /** Check if this page is current overlay and if yes
     * register it!
     */
    useEffect(() => {
        if (idx === overlayOnPageIdx) {
            const wrapper = pageWrappersRef.current.get(idx);
            setOverlayContainer(wrapper);
            setOverlayRect({
                x: (wrapper.offsetLeft + wrapper.scrollWidth) / 2 - 50,
                y: (wrapper.offsetTop + wrapper.scrollHeight) / 2 - 50,
                width: 100,
                height: 100,
            });
        }
    }, [
        idx,
        overlayOnPageIdx,
        pageWrappersRef,
        setOverlayContainer,
        setOverlayRect,
    ]);

    return (
        <div
            ref={setPageWrapperRef(idx)}
            {...props}
            className={styles.pageWrapper}
        >
            <div className={styles.page}>
                <Canvas
                    id="overlay"
                    worker={overlayWorker}
                    style={{ zIndex: 1 }}
                />
                <Canvas id="main" worker={worker} />
            </div>
        </div>
    );
}

export function _centerElementInParent(elem: HTMLElement) {
    const parent = elem.parentElement;
    if (!parent) return;

    const elemWidth = elem.scrollWidth;
    const elemHeight = elem.scrollHeight;

    const parentWidth = parent.clientWidth;
    const parentHeight = parent.clientHeight;

    const x = (parentWidth - elemWidth) / 2;
    const y = (parentHeight - elemHeight) / 2;

    elem.style.top = y + "px";
    elem.style.left = x + "px";
}
