/**
 * README:
 * -------
 *
 * The overlay "worker" is used in the main thread to render
 * the overlay canvas. This is done to avoid the overhead of
 * creating a webworker for the overlay.
 *
 * Thus it is not a real worker but a facade to the overlay
 * rendering functions.
 */
import { Socket } from "socket.io-client";

import type { ID, PageData } from "@snip/database/types";
import { BaseData, BaseView, SnipData } from "@snip/snips/general/base";
import { get_snip_from_data } from "@snip/snips/get_snip_from_data";

import { RenderWorker } from "./abc";

const debug = (...args: unknown[]) => {
    console.debug("[OverlayWorker]", ...args);
};

type RenderFunction = (ctx: CanvasRenderingContext2D) => void;

export class OverlayWorker extends RenderWorker {
    private ctx?: CanvasRenderingContext2D;
    private stack = new Set<RenderFunction>();
    private socket?: Socket;

    // Canvas size in db coordinates
    private max: [number, number] = [1400, 2000];

    public laserPointerHandler: LaserPointerHandler;
    constructor(socket: Socket) {
        super();
        this.socket = socket;

        if (!socket) {
            return;
        }

        // Attach socket handler(s)
        this.socket.on("snip:preview", this.onSnipPreview);
        this.laserPointerHandler = new LaserPointerHandler(this, socket);
    }

    public async setupCanvas(canvas: HTMLCanvasElement) {
        this.canvas = canvas;
        this.ctx = canvas.getContext("2d", {
            desynchronized: true,
        });
        debug("Setup");
    }

    public async setupPage(_page: PageData) {
        // Check if page is already setup
        if (this.current_page_id === _page.id || !this.ctx) return;
        if (this.current_page_id) {
            // Leave room
            this.socket?.emit("page:leave", this.current_page_id);
        }
        this.current_page_id = _page.id;

        // Clear stack from previews
        for (const renderfn of this.preview2render.values()) {
            this.remove(renderfn);
        }
        this.preview2render.clear();
        this.resize(this.canvas.width, this.canvas.height);

        if (!this.socket) return;

        // Join room
        return new Promise<void>((resolve, reject) => {
            this.socket.emit(
                "page:join",
                this.current_page_id,
                (success: boolean) => {
                    if (success) {
                        resolve();
                    } else {
                        reject();
                    }
                },
            );
        });
    }

    public sendPreviewData(snipData: SnipData<BaseData, BaseView>) {
        if (!this.socket) return;
        this.socket.volatile.emit("snip:preview", snipData);
    }

    private preview2render = new Map<
        number,
        (ctx: CanvasRenderingContext2D) => void
    >();
    public onSnipPreview(snipData: SnipData<BaseData, BaseView>) {
        if (!this.current_page_id || snipData.page_id !== this.current_page_id)
            return;

        // Check if snip is already rendered and remove it
        if (this.preview2render.has(snipData.id)) {
            const renderfn = this.preview2render.get(snipData.id);
            this.remove(renderfn);
        }

        // Add to stack
        const snip = get_snip_from_data(snipData);
        this.preview2render.set(snipData.id, (ctx) => {
            snip.render(ctx);
        });
        this.add(this.preview2render.get(snipData.id));
    }

    public resize(
        width: number,
        height: number,
        devicePixelRatio = 1,
        by_width = false,
    ) {
        if (!this.canvas) {
            return;
        }

        // without zoom the canvas is 1400x2000
        this.canvas.width = width * devicePixelRatio;
        this.canvas.height = height * devicePixelRatio;

        const max_db = [1400, 2000];
        let scale = height / max_db[1];
        if (by_width) {
            scale = width / max_db[0];
        }

        this.ctx.setTransform(scale, 0, 0, scale, 0, 0);
        this.max = this.canvasCoords_2_dbCoords([
            this.canvas.width,
            this.canvas.height,
        ]) as [number, number];
        this.rerender();
    }

    public rerender() {
        this.ctx.clearRect(0, 0, this.max[0], this.max[1]);
        if (this.stack.size === 0) return;
        this.stack.forEach((fn) => fn(this.ctx));
    }

    public add(fn: (ctx: CanvasRenderingContext2D) => void) {
        // Only allow function to be added once
        if (this.stack.has(fn)) return;
        this.stack.add(fn);
        this.rerender();
        debug("Added", 1);
        //print length
        debug("Stack Length", this.stack.size);
    }
    public remove = (fn: (ctx: CanvasRenderingContext2D) => void) => {
        // Get all indexes of the function in the stack
        this.stack.delete(fn);
        this.rerender();
        debug("Stack Length", this.stack.size);
    };

    public terminate() {
        this.ctx?.clearRect(0, 0, 1400, 2000);
        debug("Terminated");
    }

    /** Transform a point from canvas coordinates to
     * transformed coordinates (1400x2000).
     */
    public canvasCoords_2_dbCoords(xy: [number, number]) {
        const scale = this.canvas.height / 2000;

        const x = xy[0] / scale;
        const y = xy[1] / scale;
        return [x, y] as [number, number];
    }
    public dbCoords_2_canvasCoords(xy: [number, number]) {
        const scale = this.canvas.height / 2000;
        const x = scale * xy[0];
        const y = scale * xy[1];
        return [x, y];
    }

    public canvasCoords_2_dbCoords_het(xy: [number, number]) {
        const transform = this.ctx.getTransform().invertSelf();
        const x = transform.a * xy[0] + transform.c * xy[1] + transform.e;
        const y = transform.b * xy[0] + transform.d * xy[1] + transform.f;
        return [x, y] as [number, number];
    }

    attachElementEvents(): () => void {
        //Does nothing for the overlay worker
        return () => {};
    }

    public getScale() {
        if (!this.canvas) return 1;
        return this.canvas.height / 2000;
    }
}

/** Render handler for the laserpointer tool
 * i.e. communicate new positions and
 * render pointer events.
 */
class LaserPointerHandler {
    worker: OverlayWorker;
    socket: Socket;

    // We keep a list of received laser pointer events
    laserPointerEvents: {
        xy: [number, number];
        timestamp: number;
        socket_id: string;
        user_email: string;
    }[];

    map: Map<string, { xy: [number, number]; timestamp: number }[]>;

    constructor(worker: OverlayWorker, socket: Socket) {
        this.worker = worker;
        this.socket = socket;
        this.laserPointerEvents = [];
        this.map = new Map();
        socket.on("laserPointer:pos", this.onPos.bind(this));
    }

    private onPos(
        xy: [number, number],
        pageId: ID,
        socket_id: string,
        user_email: string,
        timestamp: number,
    ) {
        if (pageId != this.worker.current_page_id) return;

        // append to socket data
        const socket_data = this.map.get(socket_id) || [];
        socket_data.push({ xy, timestamp });
        this.map.set(socket_id, socket_data);

        // Launch render
        if (!this.renderFkt) {
            this.renderFkt = this.render.bind(this);
            this.worker.add(this.renderFkt);
            this.worker.rerender();
        }
    }

    public sendPos(xy: [number, number]) {
        if (!this.socket) return;
        this.socket.emit("laserPointer:pos", xy, this.worker.current_page_id);
    }

    renderFkt: RenderFunction;

    // Keep lines for 1s
    lineDuration = 1000;
    // Keep circles for 10s
    circleDuration = 5000;

    /** Render all laserPointer events */
    private render(ctx: CanvasRenderingContext2D) {
        const now = Date.now();

        for (const [socket_id, socket_data] of this.map.entries()) {
            // For each socket we draw a a line
            ctx.beginPath();
            ctx.strokeStyle = "rgba(255, 0, 0, 0.2)";
            ctx.lineCap = "round";
            ctx.lineWidth = 10;
            for (let i = socket_data.length - 2; i >= 0; i--) {
                ctx.lineTo(socket_data[i].xy[0], socket_data[i].xy[1]);

                // Splice after duration is up
                if (socket_data[i].timestamp < now - this.lineDuration)
                    socket_data.splice(i, 1);
            }
            ctx.stroke();

            //We also draw a bigger circle on the current position
            const last_data = socket_data.at(-1);
            if (!last_data) {
                this.map.delete(socket_id);
                continue;
            }

            ctx.beginPath();
            ctx.fillStyle = "rgba(255,0,0,0.5)";
            ctx.arc(last_data.xy[0], last_data.xy[1], 10, 0, Math.PI * 2);
            ctx.fill();

            // Splice after duration is up
            if (last_data.timestamp < now - this.circleDuration)
                socket_data.pop();
        }
        requestAnimationFrame(() => {
            if (this.map.size > 0) {
                this.worker.rerender();
            } else {
                this.worker.remove(this.renderFkt);
                this.renderFkt = null;
            }
        });
    }
}
