import { FONT_URLS } from "@snip/database/fonts/internal";
import { get_snip_from_data } from "@snip/snips";
import {
    BaseData,
    BaseSnip,
    BaseView,
    SnipData,
} from "@snip/snips/general/base";
import { ImageSnip } from "@snip/snips/general/image";

import {
    InteractionMessageEvent,
    InteractionResponseEvent,
    InteractionResponseMap,
    MessageMap,
    ResponseMap,
} from "./types";

import { Page } from "controller/page";

const debug = (...args: unknown[]) => {
    console.debug("[PageWorker]", ...args);
};

/** Basically a namespace for our worker
 * to render a single page.
 */
export class PageWorker {
    stack: Array<(ctx: OffscreenCanvasRenderingContext2D) => void> = [];
    abortController = new AbortController();

    // Max time for a full page render (ms)
    render_timeout: number = 2000;

    // The page controller might need a refactoring
    page?: Page;
    offscreenCanvas?: OffscreenCanvas;
    currentTransform?: DOMMatrix;

    /**
     * We setup a global drawing stack and an abort controller
     * to handle the draw loop.
     */
    async setupCanvas(
        data: MessageMap<"setupCanvas">,
    ): Promise<ResponseMap<"setupCanvas">> {
        this.offscreenCanvas = data.canvas;

        let lastTime = performance.now();
        let frameCount = 0;
        const ctx = this.offscreenCanvas.getContext("2d");

        // We have to opt for a draw loop
        // as firefox has some problems with dynamic
        // updates to the canvas. Therefore
        // we keep a dynamic stack of functions
        const drawLoop = (currentTime: number) => {
            //Fps counter every 2 seconds
            if (lastTime + 2000 < currentTime) {
                debug(
                    "FPS:",
                    Math.round((frameCount * 1000) / (currentTime - lastTime)),
                );
                lastTime = currentTime;
                frameCount = 0;
            }
            frameCount++;

            // Execute all functions in the stack
            // in order

            for (let i = 0; i < this.stack.length; i++) {
                const fn = this.stack[i];
                if (fn) {
                    fn(ctx);
                }
            }
            this.stack = [];
            this.currentTransform = ctx.getTransform();

            if (!this.abortController.signal?.aborted) {
                globalThis.requestAnimationFrame(drawLoop);
            }
        };

        globalThis.requestAnimationFrame(drawLoop);
    }

    /** Setup the page using page data
     * received from the main thread.
     */
    public async setupPage(
        data: MessageMap<"setupPage">,
    ): Promise<ResponseMap<"setupPage">> {
        const pageData = data.page;
        // Check if page is already set
        debug("[setupPage]", "Setting page", pageData.id);
        if (this.page && this.page.id === pageData.id) {
            debug("[setupPage]", "Page already set");
            return;
        }
        // Leave old page
        if (this.page) {
            debug("[setupPage]", "Leaving old page");
            this.page.socket.emit("page:leave", this.page.id);
        }
        this.page = new Page(pageData);
        debug("[setupPage]", "Page created");
        await this.page.init().catch((e) => {
            console.error(e);
            throw new Error("Failed to initialize page");
        });
        debug("[setupPage]", "Page initialized");
        await this.page.loadAllSnips();

        debug("[setupPage]", "Snips loaded");
        await this.page.allSnipsPrepared();
        debug("[setupPage]", "Images prepared");

        // Render the page
        // call with resize
        await this.render_page();

        // Subscribe to changes
        const l = ["insert", "remove", "placed"];
        l.forEach((event) => {
            this.page.events.addListener(event, () => {
                debug("Get Event:", event);
                this.render_page();
            });
        });
    }

    public async rerender(
        // eslint-disable-next-line @typescript-eslint/no-unused-vars
        data: MessageMap<"render">,
    ): Promise<ResponseMap<"render">> {
        return await this.render_page();
    }

    /** Perform a full render or rerender of the
     * page.
     */
    private async render_page() {
        return await new Promise<void>((resolve, reject) => {
            this.stack.push((ctx) => {
                if (this.page) {
                    const savedTransform = ctx.getTransform();
                    const transform = ctx.getTransform().invertSelf();
                    // Transform width and height to database coordinates
                    // to clear the canvas
                    const t_width =
                        transform.a * ctx.canvas.width +
                        transform.b * ctx.canvas.height +
                        transform.e;
                    const t_height =
                        transform.c * ctx.canvas.width +
                        transform.d * ctx.canvas.height +
                        transform.f;

                    ctx.clearRect(0, 0, t_width, t_height);
                    this.page.render(ctx);
                    ctx.setTransform(savedTransform);
                }
                resolve();
            });

            setTimeout(() => {
                reject(new Error("Timeout rendering page exceeded!"));
            }, this.render_timeout);
        });
    }

    /** Register fonts for the canvas/page
     * At the moment this just registers every font but
     * we might want to make this more dynamic in the future.
     */
    public async init(data: MessageMap<"init">): Promise<ResponseMap<"init">> {
        this.render_timeout = data.render_timeout;
        const promises: Promise<void>[] = [];
        for (const [url, name] of FONT_URLS) {
            promises.push(
                fetch("/fonts/" + url)
                    .then(async (response) => {
                        return response.arrayBuffer();
                    })
                    .then((data) => {
                        return new FontFace(name, data);
                    })
                    .then((font) => {
                        globalThis.fonts.add(font);
                    }),
            );
        }
        return await Promise.all(promises).then(() => {
            debug("[init]", "Fonts loaded");
            return;
        });
    }

    /** Resize the page */
    public async resize(
        data: MessageMap<"resize">,
    ): Promise<ResponseMap<"resize">> {
        if (!this.offscreenCanvas) {
            throw new Error("Canvas needs to be initialized to resize!");
        }

        //Render again
        const max_db = [1400, 2000];
        await new Promise<void>((resolve) => {
            this.stack.push((ctx) => {
                // Set the canvas size inside the render loop
                // to prevent flickering
                this.offscreenCanvas.width = data.width * data.pixelRatio;
                this.offscreenCanvas.height = data.height * data.pixelRatio;
                const scale = this.offscreenCanvas.height / max_db[1];

                ctx.setTransform(scale, 0, 0, scale, 0, 0);
                // Log performance
                this?.page?.render(ctx);
                resolve();
            });
        });
    }

    /** Shutdown the worker */
    public async stop() {
        if (this.page) {
            this.page.socket.emit("page:leave", this.page.id);
            this.page.events.removeAllListeners();
            this.page.socket.close();
        }
        this.abortController.abort();
        delete this.page;
    }

    /** Insert a snip on the page
     */
    public async insertSnip(
        data: MessageMap<"insertSnip">,
    ): Promise<ResponseMap<"insertSnip">> {
        if (!this.page) {
            throw new Error("Page needs to be initialized to insert snips!");
        }
        debug("[insertSnip]", "Inserting snip", data.snip.id);
        // Insert local
        const snip = await loadSnip(data.snip);

        this.stack.push(snip.render.bind(snip));
        this.page.add_snip(snip);

        // Insert db
        snip.socket = this.page.socket;
        snip.id = await snip.insert();
    }

    public async placeOnPage(
        data: MessageMap<"placeOnPage">,
    ): Promise<ResponseMap<"placeOnPage">> {
        if (!this.page) {
            throw new Error("Page needs to be initialized to place snips!");
        }
        const snip = await loadSnip(data.snip);
        debug("[placeOnPage]", "Placing snip", snip.id);

        this.page.add_snip(snip);

        return await new Promise<void>((resolve, reject) => {
            this.stack.push((ctx) => {
                snip.render(ctx);
                // Insert snip into database
                snip.socket = this.page.socket;
                snip.place_on_page(this.page.id);
                resolve();
            });
            setTimeout(() => {
                reject(new Error("Timeout rendering page exceeded!"));
            }, this.render_timeout);
        });
    }

    /** Setup the interaction handler */
    interactionHandler: InteractionHandler | undefined;
    public async setupInteraction(
        msg: MessageMap<"setupInteraction">,
        port: MessagePort,
    ): Promise<ResponseMap<"setupInteraction">> {
        this.interactionHandler = new InteractionHandler(this, port);
        return;
    }

    public async teardownInteraction(): Promise<
        ResponseMap<"teardownInteraction">
    > {
        this.interactionHandler.port.close();
        delete this.interactionHandler;
        return;
    }
}

class InteractionHandler {
    worker: PageWorker;
    port: MessagePort;

    constructor(worker: PageWorker, port: MessagePort) {
        this.worker = worker;
        this.port = port;

        this.port.onmessage = this.onEvent.bind(this);
    }

    /** This is triggered whenever an event is send
     * to the worker, this is a mousevent.
     */
    private async onEvent(evt: MessageEvent<InteractionMessageEvent>) {
        const { type, msg } = evt.data;

        // Base move events
        switch (type) {
            case "pointerenter":
                return this.onPointerEnter(msg);
            case "pointermove":
                return this.onPointerMove(msg);
            case "pointerleave":
                return this.onPointerLeave();
            case "dblclick":
                return this.onDblClick(msg);
            case "pointerdown":
                return this.onPointerDown(msg);
            case "pointerup":
            case "pointerout":
                console.warn("Not implemented", type);
                return;
        }
        return assertUnreachable(type);
    }

    // At the moment we only have one snip under the cursor
    // we might want to change this in the future.
    // if we allow multitouch
    snip_under_cursor: BaseSnip | null = null;

    onPointerEnter(msg: InteractionMessageEvent["msg"]) {
        //Check if a snip is under the cursor on enter
        const coords = this.getCoords(msg.mousePos);
        const snip = this.getSniptAt(coords);
        if (snip && snip.id !== this.snip_under_cursor?.id) {
            this._sendResponse("pointerenter", {
                snipData: snip.to_data(),
            });
            this.snip_under_cursor = snip;
        }
    }

    onPointerMove(msg: InteractionMessageEvent["msg"]) {
        const coords = this.getCoords(msg.mousePos);
        const snip = this.getSniptAt(coords);

        if (snip?.id !== this.snip_under_cursor?.id) {
            if (this.snip_under_cursor) {
                this._sendResponse("pointerleave", {
                    snipData: this.snip_under_cursor.to_data(),
                });
            }
            if (snip) {
                this._sendResponse("pointerenter", {
                    snipData: snip.to_data(),
                });
            }
            this.snip_under_cursor = snip;
        }
    }

    onPointerLeave() {
        if (this.snip_under_cursor) {
            this._sendResponse("pointerleave", {
                snipData: this.snip_under_cursor.to_data(),
            });
            this.snip_under_cursor = null;
        }
    }

    onDblClick(msg: InteractionMessageEvent["msg"]) {
        const coords = this.getCoords(msg.mousePos);
        const snip = this.getSniptAt(coords);
        if (snip) {
            this._sendResponse("dblclick", {
                snipData: snip.to_data(),
            });
        }
    }

    onPointerDown(msg: InteractionMessageEvent["msg"]) {
        debug("Pointer down", msg);
        const coords = this.getCoords(msg.mousePos);
        const snip = this.getSniptAt(coords);
        if (snip) {
            this._sendResponse("pointerdown", {
                snipData: snip.to_data(),
            });
        }
    }

    private getCoords(mousePos: [number, number]): [number, number] {
        const transform = DOMMatrix.fromMatrix(this.worker.currentTransform);
        transform.invertSelf();
        return [
            transform.a * mousePos[0] + transform.c * mousePos[1] + transform.e,
            transform.b * mousePos[0] + transform.d * mousePos[1] + transform.f,
        ];
    }

    private getSniptAt(coords: [number, number]): BaseSnip | null {
        if (!this.worker.page) {
            return null;
        }
        return this.worker.page.get_snip_at(coords) || null;
    }

    private _sendResponse<K extends keyof InteractionResponseMap>(
        type: K,
        data: InteractionResponseMap[K],
    ) {
        const response: InteractionResponseEvent<K> = {
            type,
            r: data,
        };
        this.port.postMessage(response);
    }
}

// eslint-disable-next-line @typescript-eslint/no-unused-vars
function assertUnreachable(x: never): never {
    throw new Error("Didn't expect to get here " + x);
}

async function loadSnip(snipData: SnipData<BaseData, BaseView>) {
    const snip = get_snip_from_data(snipData);
    if (Object.hasOwn(snip, "ready")) {
        await (snip as ImageSnip).ready;
    }
    return snip;
}
