import { HTMLProps, useEffect, useRef, useState } from "react";
import { Button, Form, InputGroup } from "react-bootstrap";
import { TbLock, TbLockOpen } from "react-icons/tb";

import { debounce } from "@snip/utils";

import { Input, InputFeedback, InputProps, ResetButton } from "./base";

import styles from "./inputs.module.scss";

/**
 * Renders a number input field.
 *
 * @param name - The name attribute of the input field.
 * @param label - The label for the input field.
 * @param props - Additional props to be passed to the input field.
 * @returns The rendered number input field.
 */
export function NumberInput({
    name = "number",
    label = "Number",
    ...props
}: InputProps) {
    return <Input name={name} label={label} type="number" {...props} />;
}

/**
 * Renders a range input component.
 *
 * @param {InputProps} props - The input component props.
 * @param {string} props.name - The name of the input.
 * @param {string} props.label - The label for the input.
 * @param {boolean} props.isInvalid - Indicates if the input is invalid.
 * @param {boolean} props.isValid - Indicates if the input is valid.
 * @param {Function} props.onReset - The callback function to reset the input.
 * @param {string} props.invalidFeedback - The feedback message for invalid input.
 * @param {string} props.validFeedback - The feedback message for valid input.
 * @returns {JSX.Element} The rendered range input component.
 */
export function RangeInput({
    name = "range",
    label = "Range",
    ...props
}: InputProps) {
    //Pop the relevant props from the props object
    //and pass the rest of the props to the input element
    const {
        isInvalid,
        isValid,
        onReset,
        invalidFeedback,
        validFeedback,
        ...rest
    } = props;

    return (
        <Input
            name={name}
            label={label}
            type="range"
            as="div"
            isInvalid={isInvalid}
            isValid={isValid}
            invalidFeedback={invalidFeedback}
            validFeedback={validFeedback}
            onReset={onReset}
        >
            <RawRangeInput {...rest} />
        </Input>
    );
}

function RawRangeInput(props: InputProps) {
    const inputRef = useRef<HTMLInputElement>(null);

    const min_size_thumb = 0.7;
    const max_size_thumb = 1.2;
    function setThumbSize(size: number) {
        const input = inputRef.current;
        if (input) {
            const scaled_size = (size - 1) / (25 - 1);
            let size_thumb =
                scaled_size * (max_size_thumb - min_size_thumb) +
                min_size_thumb;
            if (size_thumb > max_size_thumb) {
                size_thumb = max_size_thumb;
            }
            input.style.setProperty("--size-thumb", size_thumb + "em");
        }
    }

    useEffect(() => {
        setThumbSize(props.value ? parseFloat(props.value.toString()) : 0);
    }, [props.value]);

    return (
        <input
            ref={inputRef}
            type="range"
            className={"form-range " + styles.range}
            {...props}
        />
    );
}

export function RawNumberInput({ ...props }: InputProps) {
    const inputRef = useRef<HTMLInputElement>(null);

    return (
        <input
            ref={inputRef}
            type="number"
            size={3}
            min={0}
            max={25}
            // eslint-disable-next-line @typescript-eslint/no-explicit-any
            onBeforeInput={(event: any) => {
                if (!/^\d*\.?\d{0,2}$/.test(event.data)) {
                    event.preventDefault();
                }
            }}
            {...props}
        />
    );
}

export function JoinedRangeNumberInput({
    name = "range",
    label = "Range",
    isInvalid,
    isValid,
    invalidFeedback,
    validFeedback,
    onReset,
    className,
    ...props
}: InputProps) {
    return (
        <div className={className}>
            <InputGroup>
                <Form.Floating>
                    <Form.Control
                        id={name}
                        name={name}
                        placeholder=""
                        autoComplete={name}
                        as={"div"}
                        isInvalid={isInvalid}
                        isValid={isValid}
                        {...props}
                        className={styles.joinedRangeNumber}
                    >
                        <RawRangeInput {...props} />
                        <RawNumberInput {...props} />
                    </Form.Control>

                    <label htmlFor={name}>{label}</label>
                </Form.Floating>
                {onReset && <ResetButton onReset={onReset} />}
            </InputGroup>
            <InputFeedback
                isInvalid={isInvalid}
                isValid={isValid}
                invalidFeedback={invalidFeedback}
                validFeedback={validFeedback}
            />
        </div>
    );
}

export type JoinedNumbersInputProps = {
    names?: [string, string];
    label?: string;
    values?: [number, number];
    defaultValues?: [number, number];
    onReset?: () => void;
    onChange?: (first?: number, sec?: number) => void;
} & Omit<InputProps, "onChange">;

export function JoinedNumbersInput({
    names = ["number1", "number2"],
    label = "JoinedNumbers",
    className,
    onReset,
    onChange,
    values,
    defaultValues,
    ...props
}: JoinedNumbersInputProps) {
    let cn = styles.joinedNumbers;
    if (className) {
        cn += " " + className;
    }

    return (
        <div className={cn}>
            <InputGroup>
                <Form.Floating>
                    <Form.Control
                        id={names[0] + names[1]}
                        as={"div"}
                        className={styles.joinedNumbers}
                    >
                        <RawNumberInput
                            id={names[0]}
                            name={names[0]}
                            value={values ? values[0] : undefined}
                            defaultValue={
                                defaultValues ? defaultValues[0] : undefined
                            }
                            onChangeCapture={(e) => {
                                if (onChange) {
                                    onChange(
                                        parseFloat(e.target.value),
                                        undefined,
                                    );
                                }
                            }}
                            {...props}
                        />
                        <RawNumberInput
                            id={names[1]}
                            name={names[1]}
                            value={values ? values[1] : undefined}
                            defaultValue={
                                defaultValues ? defaultValues[1] : undefined
                            }
                            onChangeCapture={(e) => {
                                if (onChange) {
                                    onChange(
                                        undefined,
                                        parseFloat(e.target.value),
                                    );
                                }
                            }}
                            {...props}
                        />
                    </Form.Control>
                    <label>{label}</label>
                </Form.Floating>
                {onReset && <ResetButton onReset={onReset} />}
            </InputGroup>
        </div>
    );
}

export interface LinkedNumbersInputProps extends HTMLProps<HTMLDivElement> {
    labels: [string, string];
    names: [string, string];
    defaultValues: [number, number];
    handleChange: (values: [number, number], locked: boolean) => void;
    step?: number;
}

/** Inputs which allow to
 * lock the aspect ratio
 * between two linked number inputs
 */
export function LinkedNumbersInput({
    names,
    labels,
    className,
    defaultValues,
    step,
    handleChange,
    ...props
}: LinkedNumbersInputProps) {
    const [locked, setLocked] = useState(false);
    const [values, setValues] = useState(defaultValues);

    const debouncedSetValues = useRef(
        debounce((values: [number, number]) => {
            setValues(values);
        }, 100),
    ).current;

    let cn = styles.linkedNumbers;
    if (className) {
        cn += " " + className;
    }

    useEffect(() => {
        // Round to max 2 decimal places
        const n = [
            Math.round(defaultValues[0] * 100) / 100,
            Math.round(defaultValues[1] * 100) / 100,
        ];

        debouncedSetValues(n);
    }, [debouncedSetValues, defaultValues]);

    return (
        <div className={cn} {...props}>
            <div className="d-flex flex-column gap-2">
                <NumberInput
                    id={names[0]}
                    name={names[0]}
                    label={labels[0]}
                    value={values[0]}
                    onInput={(e) => {
                        const n: [number, number] = [
                            parseFloat(e.target.value),
                            locked
                                ? (parseFloat(e.target.value) * values[1]) /
                                  values[0]
                                : values[1],
                        ];

                        // Round to max 2 decimal places
                        n[0] = Math.round(n[0] * 100) / 100;
                        n[1] = Math.round(n[1] * 100) / 100;

                        setValues(n);

                        handleChange(n, locked);
                    }}
                    step={step}
                />
                <NumberInput
                    id={names[1]}
                    name={names[1]}
                    label={labels[1]}
                    value={values[1]}
                    onInput={(e) => {
                        const n: [number, number] = [
                            locked
                                ? (parseFloat(e.target.value) * values[0]) /
                                  values[1]
                                : values[0],
                            parseFloat(e.target.value),
                        ];

                        // Round to max 2 decimal places
                        n[0] = Math.round(n[0] * 100) / 100;
                        n[1] = Math.round(n[1] * 100) / 100;

                        setValues(n);

                        handleChange(n, locked);
                    }}
                    step={step}
                />
            </div>
            <div className={styles.lock}>
                <Button variant="none" onClick={() => setLocked(!locked)}>
                    {locked ? <TbLock /> : <TbLockOpen />}
                </Button>
            </div>
        </div>
    );
}
