import EventEmitter from "node:events";

import { Socket } from "socket.io-client";

import { BaseData, BaseView, PageData } from "@snip/database/types";
import { LocalPage } from "@snip/render/localPage";
import { get_snip_from_data_allow_unknown } from "@snip/snips";
import { ArraySnip } from "@snip/snips/general/array";
import type { BaseSnip, SnipData } from "@snip/snips/general/base";

import { getSocket } from "components/context/socket";

/** A page is able to render a page and retrieve data from the server
 * using a socket connection.
 */
export class Page extends LocalPage {
    public socket?: Socket;

    /** Event emitter for all changes to the page.
     *
     * The idea is to e.g. subscribe to the change event and
     * rerender the page.
     */
    public events = new EventEmitter();

    constructor({
        id,
        book_id,
        page_number,
        created = new Date(),
        last_updated = new Date(),
        background_type_id = null,
        snips = [],
    }: PageData & { snips?: BaseSnip[] }) {
        super({
            id,
            book_id,
            page_number,
            created,
            last_updated,
            background_type_id,
            snips,
        });
    }

    async init(): Promise<void> {
        this.socket = await getSocket(this.book_id);
        if (!this.socket) {
            throw new Error("Socket not initialized");
        }
        if (!this.socket.connected) {
            await this.socket.connect();
        }
        await this.joinRoom();
        await this.subscribeToChanges();
    }

    /** Get the preview image of the page */
    public async getPreviewImageBuffer(): Promise<Buffer> {
        return await new Promise<Buffer>((resolve) => {
            this.socket.emit(
                "page:getPreviewImage",
                this.id,
                (buffer: Buffer) => {
                    resolve(buffer);
                },
            );
        });
    }

    /** Join room for the page */
    private async joinRoom(): Promise<void> {
        if (!this.socket) {
            return;
        }
        return await new Promise<void>((resolve, reject) => {
            this.socket.emit("page:join", this.id, (success: boolean) => {
                if (success) {
                    resolve();
                } else {
                    reject();
                }
            });
        });
    }

    /** Use socket to load all snips */
    public async loadAllSnips(): Promise<void> {
        if (!this.socket) {
            throw new Error("Socket not initialized");
        }
        return new Promise<void>((resolve) => {
            this.socket.emit(
                "page:getSnips",
                this.id,
                (snips: SnipData<BaseData, BaseView>[]) => {
                    this.set_snips(
                        snips.map((snip) =>
                            get_snip_from_data_allow_unknown(snip),
                        ),
                    );
                    resolve();
                },
            );
        });
    }

    /** Images need
     * to be loaded before they can be rendered.
     */
    public async allSnipsPrepared(): Promise<boolean> {
        const snips_with_ready = this.snips.filter((snip) => {
            return Object.hasOwn(snip, "ready");
        });
        return Promise.all(snips_with_ready.map((snip) => snip.ready)).then(
            () => true,
        );
    }

    /** Subscribe to changes */
    private subscribeToChanges(): void {
        if (!this.socket) {
            throw new Error("Socket not initialized");
        }
        this.socket.on(
            "snip:insert",
            (newSnip: SnipData<BaseData, BaseView>) => {
                // Check if snip is on this page
                if (newSnip.page_id === this.id) {
                    this.add_snip(get_snip_from_data_allow_unknown(newSnip));
                    this.events.emit("insert", newSnip);
                }
            },
        );

        this.socket.on(
            "snip:update",
            (updatedSnip: SnipData<BaseData, BaseView>) => {
                // Check if snip is on this page
                if (updatedSnip.page_id === this.id) {
                    this.update_snip(
                        get_snip_from_data_allow_unknown(updatedSnip),
                    );
                    this.events.emit("update", updatedSnip);
                }
            },
        );

        this.socket.on("snip:remove", (snipId: number) => {
            // Check if snip is on this page
            const success = this.remove_snip(snipId);
            if (success) {
                this.events.emit("remove", snipId);
            }
        });

        this.socket.on("snip:placed", (pageId: number, snipId: number) => {
            // Check if snip is on this page
            if (pageId === this.id) {
                // Check if array snip and if so, update the snip
                const snip = this.snips.find((s) => s.id === snipId);
                if (snip && (snip as ArraySnip).snips) {
                    // find all snips in array and remove them
                    const array_snip = snip as ArraySnip;
                    const snips_to_remove = array_snip.snips.map((s) => s.id);
                    // find in snips
                    this.snips = this.snips.filter(
                        (s) => !snips_to_remove.includes(s.id),
                    );
                }
                this.events.emit("placed", snipId);
            }
        });
    }
    static from_data(data: PageData): Page {
        return new Page({
            ...data,
        });
    }
}
