"use client";
import { useCallback, useEffect, useState } from "react";

/** Mouse delta handler returns the delta of the mouse position from the last frame movements.
 *
 * It is possible to automatically attach the event listeners to a specific element by passing a ref to the hook.
 *
 *
 * Returns three handlers:
 * - mouseDown: sets the mouse down state to true
 * - mouseUp: sets the mouse down state to false
 * - mouseMove: sets the mouse position state
 * - mouse: Object which contains the mouse position and delta
 *
 */
const useMouseDelta = (ref: React.RefObject<HTMLElement> = undefined) => {
    // state to save the current mouse positions between move events
    const [mouse, setMouse] = useState<{
        prev_x: number;
        prev_y: number;
        current_x: number;
        current_y: number;
    }>(undefined);

    const mouseMove = useCallback((e) => {
        e.preventDefault();
        // Compute difference between initial and current mouse position
        setMouse((prev) => {
            return {
                prev_x: prev.current_x,
                prev_y: prev.current_y,
                current_x: e.clientX,
                current_y: e.clientY,
            };
        });
    }, []);

    const mouseUp = useCallback(
        (e) => {
            e.preventDefault();
            // Remove event listeners
            window.removeEventListener("pointermove", mouseMove);
            window.removeEventListener("pointerup", mouseUp);
        },
        [mouseMove]
    );

    const mouseDown = useCallback(
        (e) => {
            e.preventDefault();
            // Set initial mouse position
            setMouse({
                prev_x: e.clientX,
                prev_y: e.clientY,
                current_x: e.clientX,
                current_y: e.clientY,
            });

            // Add event listeners
            window.addEventListener("pointermove", mouseMove);
            window.addEventListener("pointerup", mouseUp);
        },
        [mouseMove, mouseUp]
    );

    useEffect(() => {
        if (!ref || !ref.current) return;

        const el = ref.current;

        el.addEventListener("pointerdown", mouseDown);

        return () => {
            el.removeEventListener("pointerdown", mouseDown);
        };
    }, [mouseDown, ref]);

    return {
        mouseDown,
        mouseUp,
        mouseMove,
        mouse,
    };
};

export default useMouseDelta;
