import request from "supertest";
import { server } from "../server";
import { config } from "@snip/utils/config";
import { mockConfig } from "../../test_setup/mockConfig";

describe("/render", () => {
    beforeAll(() => {
        server.listen(3000);
    });

    afterAll(() => {
        server.close();
    });

    it("should return 400 if no search params are provided", async () => {
        await request(server).get("/render").expect(400);
    });

    it("should return 400 if invalid search params are provided", async () => {
        await request(server)
            .get("/render")
            .query({ page_id: "invalid" })
            .expect(400);

        await request(server)
            .get("/render")
            .query({ page_id: 1, width: "invalid" })
            .expect(400);

        await request(server)
            .get("/render")
            .query({ page_id: 1, width: 200, format: "invalid" })
            .expect(400);

        await request(server)
            .get("/render")
            .query({
                page_id: 1,
                width: 200,
                format: "jpeg",
                encoding: "invalid",
            })
            .expect(400);
    });

    it("should return 400 if a rerender is forced by a non-api user", async () => {
        mockConfig.force_api_session = false;
        await request(server)
            .get("/render")
            .query({
                page_id: 1,
                width: 200,
                format: "jpeg",
                force_render: true,
            })
            .expect(400)
            .expect("Force render is not allowed for user requests");
        mockConfig.force_api_session = true;
    });

    it("should return 401 if no session is provided", async () => {
        mockConfig.force_api_session = false;
        mockConfig.session = false;
        await request(server)
            .get("/render")
            .query({ page_id: 1, width: 200, format: "jpeg" })
            .expect(401);
        mockConfig.force_api_session = false;
        mockConfig.session = true;
    });

    it("query params", async () => {
        const res = await request(server)
            .get("/render")
            .query({ page_id: 1, width: 200, format: "jpeg" })
            .expect("Content-Type", "image/jpeg")
            .expect("Content-Encoding", "gzip")
            .expect(200);

        await request(server)
            .get("/render")
            .query({ page_id: 1, width: 1400, format: "pdf", encoding: "none" })
            .expect("Content-Type", "application/pdf")
            .expect(200);
    });

    it("body params", async () => {
        await request(server)
            .post("/render")
            .send({ page_id: 1, width: 200, format: "jpeg" })
            .expect("Content-Type", "image/jpeg")
            .expect("Content-Encoding", "gzip")
            .expect(200);

        await request(server)
            .post("/render")
            .send({ page_id: 1, width: 200, format: "jpeg", encoding: "none" })
            .expect("Content-Type", "image/jpeg")
            .expect(200);
    });

    it("www-form-urlencoded", async () => {
        const res = await request(server)
            .post("/render")
            .send("page_id=1&width=200&format=jpeg")
            .expect("Content-Type", "image/jpeg")
            .expect("Content-Encoding", "gzip")
            .expect(200);
    });
});

describe("/render/*.png|jpeg|webp", () => {
    it("should return 404 not found if not an id ", async () => {
        const endings = config.RENDER.FORMATS.map((f) => f.format);
        for (const ending of endings) {
            await request(server)
                .get("/render/test." + ending)
                .expect(404)
                .expect("Not found");
        }
    });

    it("should return 403 if id not found", async () => {
        const endings = config.RENDER.FORMATS.map((f) => f.format);
        const invalid_id = 9999;
        for (const ending of endings) {
            const res = await request(server)
                .get(`/render/${invalid_id}.${ending}?w=1200`)
                .expect(403);
        }
    });

    it("should allow for 'w' and 'width' query params", async () => {
        const endings = config.RENDER.FORMATS.map((f) => f.format);
        const id = 1;
        for (const ending of endings) {
            await request(server)
                .get(`/render/${id}.${ending}?w=12`)
                .expect(200);

            await request(server)
                .get(`/render/${id}.${ending}?width=12`)
                .expect(200);
        }
    });
});

describe("any /", () => {
    it("should return 404 if wrong endpoint", async () => {
        await request(server).get("/").expect(404);

        await request(server).get("/wrong").expect(404);
    });
});

describe("any /rerender/:page_id", () => {
    it("should return 200 on simple request", async () => {
        mockConfig.force_api_session = true;
        await request(server).get("/rerender/1").expect(200);
        mockConfig.force_api_session = false;
    });
});

describe("any /render/book_preview/:book_id", () => {
    it("should return 200 on simple request", async () => {
        const endings = config.RENDER.FORMATS.map((f) => f.format);
        for (const ending of endings) {
            await request(server)
                .get(`/render/book_preview/1.${ending}`)
                .expect(200);
        }
    });
});
