import { config } from "@snip/utils/config";

export interface RenderParams {
    encoding: "gzip" | "none";
    format: "png" | "webp" | "jpeg" | "pdf";
    width: number;
    page_id: number;
    force_render: boolean;
    return_stream: boolean;
}

/** Validate the given search params
 *  and return the parsed search params
 *  if they are valid.
 *
 * If the search params are invalid, the function
 * throws an error with a message which can be
 * displayed to the user.
 *
 *
 */
export function validate_renderParams(params_raw: {
    encoding?: string;
    page_id?: string;
    format?: string;
    width?: string;
    w?: string;
    return_stream?: string | boolean;
    force_render?: string | boolean;
}): RenderParams {
    if (params_raw === undefined) {
        throw new Error("You have to provide the 'page_id' as a parameter!");
    }

    // Parse encoding
    const encoding = params_raw["encoding"] as "gzip" | "none";
    if (encoding) {
        if (encoding !== "gzip" && encoding !== "none") {
            throw new Error(
                "Invalid 'encoding' query parameter (not 'gzip' or 'none')!",
            );
        }
    }

    // Parse page_id
    // only one can be given
    const page_id = params_raw["page_id"];
    if (!page_id) {
        throw new Error("You have to provide the 'page_id' as a parameter!");
    } else {
        // Check if number is valid
        if (isNaN(parseInt(page_id)) || parseInt(page_id) < 1) {
            throw new Error(
                "Invalid 'page_id' parameter (not a number or smaller 1)!",
            );
        }
    }

    // Parse format
    const format = parseFormat(params_raw["format"]);

    // Parse width
    const width = parseWidth(params_raw["width"] || params_raw["w"], format);

    // Force render
    const force_render = params_raw["force_render"];
    if (force_render) {
        if (
            force_render !== "true" &&
            force_render !== "false" &&
            (force_render as boolean) !== true &&
            (force_render as boolean) !== false
        ) {
            throw new Error(
                "Invalid 'force_render' query parameter (not 'true' or 'false')!",
            );
        }
    }

    let return_stream = params_raw["return_stream"];
    if (return_stream) {
        if (
            return_stream !== "true" &&
            return_stream !== "false" &&
            (return_stream as boolean) !== true &&
            (return_stream as boolean) !== false
        ) {
            throw new Error(
                "Invalid 'return_stream' query parameter (not 'true' or 'false')!",
            );
        }
    } else {
        return_stream = true;
    }

    return {
        encoding: encoding ? encoding : "gzip",
        width: width,
        format: format,
        page_id: parseInt(page_id),
        force_render: force_render
            ? force_render === "true" || force_render === true
            : false,
        return_stream: return_stream
            ? return_stream === "true" || return_stream === true
            : true,
    };
}

function parseWidth(w: string | undefined, format: string): number {
    let width: number;
    // Check if number is valid
    if (w) {
        if (isNaN(parseInt(w))) {
            throw new Error("Invalid 'width' query parameter (not a number)!");
        }
        width = parseInt(w);

        const allowed_widths = config.RENDER.FORMATS.find(
            (f) => f.format === format,
        )?.widths;
        width = allowed_widths.reduce(function (prev, curr) {
            return Math.abs(curr - width) < Math.abs(prev - width)
                ? curr
                : prev;
        });
    } else {
        width = config.RENDER.FORMATS.find((f) => f.format === format)
            ?.widths[0];
    }
    return width;
}

function parseFormat(f: string | undefined): RenderParams["format"] {
    let format: string;
    if (f) {
        const allowed_formats = config.RENDER.FORMATS.map((f) => f.format);
        if (!allowed_formats.includes(f)) {
            throw new Error(
                "Invalid 'format' query parameter! Allowed formats are: " +
                    allowed_formats.join(", "),
            );
        }
        format = f;
    } else {
        format = config.RENDER.FORMATS[0].format;
    }
    return format as RenderParams["format"];
}

export type SelectedPage = [number, number] | number;
export type PageNumberLocation = "none" | "br" | "tl" | "tr" | "bl" | "custom";

/** Validates and parses the selectedPages string
 *
 * throws an error if the string is invalid
 *
 */
export function parseSelectedPages(selectedPages: string): SelectedPage[] {
    const pages: SelectedPage[] = [];

    // Split by comma and parse items
    const items = selectedPages.split(",");

    for (const item of items) {
        if (item === "") {
            continue;
        }

        // Check if item is a range
        if (item.includes("-")) {
            const [start, end] = item.split("-");

            // check if start and end are numbers
            if (
                isNaN(parseInt(start)) ||
                isNaN(parseInt(end)) ||
                !/^\d+$/.test(start) ||
                !/^\d+$/.test(end)
            ) {
                throw new Error(`Invalid range in page numbers! "${item}"`);
            }

            // Check if start is smaller than end
            if (parseInt(start) > parseInt(end)) {
                throw new Error(`Invalid range in page numbers! "${item}"`);
            }

            // Start bigger than 0
            if (parseInt(start) < 1) {
                throw new Error(
                    `First item in range has to be bigger than 0! "${item}"`,
                );
            }
            pages.push([parseInt(start), parseInt(end)]);
        }

        // Check if item is a single page
        else {
            // check if item is a number
            if (isNaN(parseInt(item)) || !/^\d+$/.test(item)) {
                throw new Error(`Invalid page number! "${item}"`);
            }

            // Check if item is bigger than 0
            if (parseInt(item) < 1) {
                throw new Error(
                    `Page number has to be bigger than 0! "${item}"`,
                );
            }
            pages.push(parseInt(item));
        }
    }

    return pages;
}
