import service from "@snip/database";
import { ID, PageData } from "@snip/database/types";
import { BaseData, BaseView, SnipData } from "@snip/snips/general/base";

import { type SocketData } from "../../types";
import { addAllowedPage, allowed, allowedPage } from "../../utils/auth";
import { logWS } from "../../utils/logging";

export function setupBookPrefix({ socket, server }: SocketData) {
    /** Called on book init to get all page data
     *  - this does not transfer snip data
     */
    socket.on(
        "book:getPages",
        async (callback: (pages: PageData[]) => void) => {
            // Check if user has permission to view the page
            if (!allowed(socket)) {
                logWS(socket, "book:getPages", "No permission");
                return socket.emit(
                    "error",
                    "You do not have access to view this book",
                );
            }
            logWS(socket, "book:getPages", "Get pages");

            const pages = await service.page.getByBookId(socket.book_id);
            callback?.(pages);
        },
    );

    /** Insert a new page into the database
     */
    socket.on(
        "book:newPage",
        async (
            pageData: Omit<PageData, "id">,
            callback: (pageData: PageData) => void,
        ) => {
            // Check if user has permission to write to the page
            if (!allowed(socket, true, true, false)) {
                logWS(socket, "book:newPage", "No permission");
                return socket.emit(
                    "error",
                    "You do not have access to write to this book. Could be finished!",
                );
            }
            logWS(socket, "book:newPage", "New page");

            // Add the page to the database
            const inserted_page = await service.page.insert(pageData);
            // Get default background
            if (inserted_page.background_type_id === undefined) {
                inserted_page.background_type_id =
                    await service.backgroundType.getSelectedByBookId(
                        inserted_page.book_id,
                    );
            }

            // Add the page to the allowedPages of all sockets
            // in the same book
            addAllowedPage(server, inserted_page.book_id, inserted_page.id);
            // emit to all other clusternodes
            await server
                .of("/server")
                .serverSideEmitWithAck(
                    "book:newPage:sync_allowedPages",
                    inserted_page.book_id,
                    inserted_page.id,
                )
                .catch((e) => {
                    console.error(e);
                });

            // Emit to clients
            callback?.(inserted_page);
            socket.broadcast.emit("book:newPage", inserted_page);
        },
    );

    socket.on(
        "book:deletePage",
        async (page_id: number, callback?: (page_id: number) => void) => {
            // Check if user has permission to write to the page
            if (!allowedPage(socket, page_id, true, true, true)) {
                logWS(socket, "book:deletePage", "No permission", page_id);
                return socket.emit(
                    "error",
                    "You do not have access to delete from this book",
                );
            }

            logWS(socket, "book:deletePage", "Delete page " + page_id);

            // Delete the page from the database
            await service.page.delete(page_id);
            callback?.(page_id);

            // Emit to clients
            socket.broadcast.emit("book:deletePage", page_id);
        },
    );

    socket.on(
        "book:updatePage",
        async (pageData: PageData, callback: (pageData: PageData) => void) => {
            // Check if user has permission to write to the page
            if (!allowedPage(socket, pageData.id, true, true, false)) {
                logWS(socket, "book:updatePage", "No permission", pageData.id);
                return socket.emit(
                    "error",
                    "You do not have access to write to this book. Could be finished!",
                );
            }
            logWS(socket, "book:updatePage", "Update page " + pageData.id);
            // Update the page in the database
            const updated_page = await service.page.update(pageData);
            callback?.(updated_page);

            // Emit to clients
            socket.broadcast.emit("book:updatePage", updated_page);
        },
    );

    /** Called on book init to get all not yet on pages placed snips
     *
     * @deprecated
     */
    socket.on(
        "book:getQueuedSnips",
        (callback: (snipsData: SnipData<BaseData, BaseView>[]) => void) => {
            // Check if user has permission to view the page
            if (!allowed(socket)) {
                logWS(socket, "book:getQueuedSnips", "No permission");
                return socket.emit(
                    "error",
                    "You do not have access to view this book",
                );
            }
            logWS(socket, "book:getQueuedSnips", "Get queued snips");
            getQueuedSnips(socket.book_id, callback);
        },
    );

    /** Insert a snip directly via the
     * socket connection
     */
    socket.on(
        "book:newQueuedSnip",
        (snipData: SnipData<BaseData, BaseView>) => {
            // Check if user has permission to write to the page
            if (!allowed(socket, true, true)) {
                logWS(
                    socket,
                    "book:newQueuedSnip",
                    "No permission",
                    snipData.id,
                );
                return socket.emit(
                    "error",
                    "You do not have access to write to this book. Could be finished!",
                );
            }

            // insert if applicable
            // TODO

            logWS(
                socket,
                "book:newQueuedSnip",
                "New queued snip " + snipData.id,
            );
            // Emit to clients
            socket.broadcast.emit("book:newQueuedSnip", snipData);
        },
    );
    socket.on("book:emit:newQueuedSnip", async (snip_id: ID) => {
        // Check if user has permission to write to the page
        if (!allowed(socket, true, true)) {
            logWS(socket, "book:emit:NewQueuedSnip", "No permission", snip_id);
            return socket.emit(
                "error",
                "You do not have access to write to this book. Could be finished!",
            );
        }
        logWS(socket, "book:emit:NewQueuedSnip", "New queued snip " + snip_id);

        // Get snip data
        const snipData = await service.snip.getById(snip_id);

        // Emit to clients
        socket.broadcast.emit("book:newQueuedSnip", snipData);
    });

    socket.on("book:deleteQueuedSnip", (snip_id: number) => {
        // Check if user has permission to write to the page
        if (!allowed(socket, true, true, true)) {
            logWS(socket, "book:deleteQueuedSnip", "No permission", snip_id);
            return socket.emit(
                "error",
                "You do not have access to delete from this book",
            );
        }
        // TODO
        //await service.snip.deleteQueuedSnip(snip_id);

        logWS(socket, "book:deleteQueuedSnip", "Delete queued snip " + snip_id);
        // Emit to clients
        socket.broadcast.emit("book:deleteQueuedSnip", snip_id);
    });
}

/** Get all not yet placed snips for a book
 */
async function getQueuedSnips(
    book_id: number,
    callback: (snipsData: SnipData<BaseData, BaseView>[]) => void,
) {
    // Get all not placed snips by book id
    const snipsData = await service.snip.getQueuedByBookId(book_id);

    // Map the snips to data only for smaller transfer size
    callback?.(snipsData);
}
