import SQLService from "@snip/database";
import type { BaseData, BaseView, SnipData } from "@snip/snips/general/base";

import { type SocketData } from "../../types";
import { allowedPage } from "../../utils/auth";
import { logWS } from "../../utils/logging";

export function setupPagesPrefix({ socket }: SocketData) {
    /** On join of page check if the
     * page exists and if the user has access
     * as in theory the user could use the
     * client to join any page
     */
    socket.on(
        "page:join",
        async (page_id: number, callback: (success: boolean) => void) => {
            // No permission
            if (!allowedPage(socket, page_id)) {
                logWS(socket, "page:join", "No permission", page_id);
                return callback?.(false);
            }

            // Allowed
            logWS(socket, "page:join", "Join page " + page_id);
            await socket.join("page-" + String(page_id));
            callback?.(true);
        },
    );

    /** Leave the page
     * @param page_id - The ID of the page to leave
     */
    socket.on("page:leave", (page_id: number) => {
        logWS(socket, "page:leave", "Leave page " + page_id);
        socket.leave("page-" + String(page_id));
    });

    /** Get all snips of a page
     * @param page_id - The ID of the page to get snips for
     */
    socket.on(
        "page:getSnips",
        async (
            page_id: number,
            callback: (snips: SnipData<BaseData, BaseView>[]) => void,
        ) => {
            if (!allowedPage(socket, page_id)) {
                logWS(socket, "page:getSnips", "No permission", page_id);
                return callback?.([]);
            }

            const snips = await SQLService.snip.getByPageId(page_id);
            callback?.(snips);
        },
    );
}
