import { Namespace } from "socket.io";

import { ID } from "@snip/database/types";

import { RemoteSocket, SocketData, User } from "../../types";
import { logWS } from "../../utils/logging";

export function setupUsersPrefix({ socket, server }: SocketData) {
    /** Emit a user connect event if
     * the user freshly connects to the server
     * i.e. if this is the first socket connection
     * from the user
     */
    server
        .of("book-" + socket.book_id)
        .fetchSockets()
        .then((sockets) => {
            sockets = sockets.filter((s) => s.data.id === socket.data.id);
            if (sockets.length === 1) {
                //Emit user connect event
                const user = {
                    id: socket.data.id,
                    email: socket.data.email,
                    page_ids: Array.from(roomsToPageIds(socket.rooms).values()),
                };
                socket.volatile.emit("users:connect", user);
            }
        });

    /** Allow clients to request all
     * online users
     */
    socket.on("users:getOnline", async (callback: (users: User[]) => void) => {
        logWS(socket, "Get online users");
        const sockets = await server
            .of("book-" + socket.book_id)
            .fetchSockets();

        let users: (User & { self?: boolean })[] = socketsToUsers(sockets);

        users = users.map((user) => {
            if (user.id === socket.data.id) {
                return {
                    ...user,
                    self: true,
                };
            }
            return user;
        });

        callback?.(users);
    });

    /** Emit proper user disconnect
     * event when user disconnects from all sockets
     */
    socket.on("disconnect", async () => {
        // Check if the user has disconnected with all sockets
        const sockets = await server
            .of("book-" + socket.book_id)
            .fetchSockets();

        for (const s of sockets) {
            if (s.data.id === socket.data.id) {
                return;
            }
        }
        logWS(socket, "User disconnected");
        socket.volatile.emit("users:disconnect", socket.data.id);
    });
}

function socketsToUsers(sockets: RemoteSocket[]): User[] {
    const userMap: Map<ID, User> = new Map();

    sockets.forEach((socket) => {
        if (!socket.data.id) {
            return;
        }
        if (!userMap.has(socket.data.id)) {
            // Add user to map
            userMap.set(socket.data.id, {
                id: socket.data.id,
                email: socket.data.email || "?",
                page_ids: roomsToPageIds(socket.rooms),
            });
        } else {
            //Update page_ids
            const user = userMap.get(socket.data.id);
            if (user) {
                user.page_ids = new Set([
                    ...user.page_ids,
                    ...roomsToPageIds(socket.rooms),
                ]);
            }
        }
    });

    let users = Array.from(userMap.values());
    // map page ids set to array
    users = users.map((user) => ({
        ...user,
        page_ids: Array.from(user.page_ids.values()),
    }));

    // Map set to array
    return users;
}

export async function getAllSocketsById(
    namespace: Namespace,
    id: ID,
): Promise<RemoteSocket[]> {
    const sockets = await namespace.fetchSockets();
    return sockets.filter((s) => s.data.id === id);
}

function roomsToPageIds(rooms: Set<string>): Set<ID> {
    const pageIds: Set<ID> = new Set();
    rooms.forEach((room) => {
        const pageId = room.split("-")[1];
        if (pageId && !isNaN(parseInt(pageId))) {
            pageIds.add(parseInt(pageId));
        }
    });
    return pageIds;
}
