import { BookService } from "../book";
import {
    BookData,
    BookDataResolved,
    BookOwner,
    ENTITY_GROUP,
    ENTITY_USER,
    Entity,
    ID,
    PartialBy,
} from "../../types";
import { pool_data, pool_permissions } from "../../sql.connection";
import { NotFoundError } from "../../errors";

describe("BookService", () => {
    let bookService: BookService;
    let consoleSpy;
    let user_ids: ID[] = [];
    let group_ids: ID[] = [];
    let default_book: PartialBy<BookData, "id">;

    beforeEach(() => {
        if (typeof consoleSpy === "function") {
            consoleSpy.mockRestore();
        }
        consoleSpy = jest.spyOn(console, "warn").mockImplementation();
    });

    beforeAll(async () => {
        bookService = new BookService();

        // Insert user and group for testing otherwise
        // the foreign key constraints will fail

        await pool_permissions
            .query(
                "INSERT INTO users (`email`, `emailValidated`, `hashedPassword`) VALUES ('test@book.de', 1, 'test')",
            )
            .then((res) => {
                user_ids.push(res.insertId);
            });
        await pool_permissions
            .query(
                "INSERT INTO users (`email`, `emailValidated`, `hashedPassword`) VALUES ('test2@book.de', 1, 'test')",
            )
            .then((res) => {
                user_ids.push(res.insertId);
            });
        await pool_permissions
            .query(
                "INSERT INTO groups (`name`, `description`) VALUES ('book', 'test')",
            )
            .then((res) => {
                group_ids.push(res.insertId);
            });

        default_book = {
            title: "Test.book",
            comment: "This is a test book",
            finished: null,
            owner_user_id: user_ids[0],
            owner_group_id: null,
            cover_page_id: null,
            // Placeholder date to check if it is updated
            created: new Date(),
            last_updated: new Date(),
        };

        return;
    });

    afterAll(async () => {
        await pool_data.query("DELETE FROM books where title LIKE '%.book%'");
        await pool_permissions.query("DELETE FROM groups where name = 'book'");
        await pool_permissions.query(
            "DELETE FROM users where email LIKE '%book%'",
        );
    });

    describe("insert", () => {
        it("should insert a new book into the database", async () => {
            const result: BookData = await bookService.insert(
                structuredClone(default_book),
            );

            expect_same(result, default_book);

            // Only with title and owner
            const result2: BookData = await bookService.insert({
                title: "test2.book",
                owner_group_id: group_ids[0],
            });
            expect(result2.title).toBe("test2.book");
        });
    });

    describe("getById", () => {
        it("should return the same book that was inserted", async () => {
            const book: BookData = await bookService.insert(
                structuredClone(default_book),
            );

            const result: BookData = await bookService.getById(book.id);

            expect_same(result, book);
            expect(result.id).toBe(book.id);
        });

        it("should throw an error if the book does not exist", async () => {
            await expect(bookService.getById(-1)).rejects.toThrow(
                NotFoundError,
            );
        });
    });

    describe("getByIds", () => {
        it("should return the same books that were inserted", async () => {
            const book1: BookData = await bookService.insert(
                structuredClone(default_book),
            );
            const book2: BookData = await bookService.insert(
                structuredClone(default_book),
            );

            const result: BookData[] = await bookService.getByIds([
                book1.id,
                book2.id,
            ]);

            expect(result.length).toBe(2);
            expect_same(result[0], book1);
            expect_same(result[1], book2);
        });

        it("should gracefully handle empty arrays", async () => {
            const result: BookData[] = await bookService.getByIds([]);
            expect(result.length).toBe(0);
        });

        it("should return an partial array if any of the books do not exist", async () => {
            const books = await bookService.getByIds([88888, 99999]);
            expect(books.length).toBe(0);

            const book1: BookData = await bookService.insert(
                structuredClone(default_book),
            );
            const books2 = await bookService.getByIds([book1.id, 99999]);
            expect(books2.length).toBe(1);
            // Expect warn
            expect(consoleSpy).toHaveBeenCalled();
        });
    });

    describe("getByIdResolved", () => {
        it("should return the same book that was inserted", async () => {
            const book: BookData = await bookService.insert(
                structuredClone(default_book),
            );

            const result: BookDataResolved = await bookService.getByIdResolved(
                book.id,
            );

            expect(result.id).toBe(book.id);
            expect(result.owner_name).toBe("test@book.de");
            expect(result.owner_type).toBe("user");
            expect(result.owner_id).toBe(user_ids[0]);
        });

        it("should throw an error if the book does not exist", async () => {
            await expect(bookService.getByIdResolved(-1)).rejects.toThrow(
                NotFoundError,
            );
        });
    });

    describe("getByIdsResolved", () => {
        it("should return the same books that were inserted", async () => {
            const book1: BookData = await bookService.insert(
                structuredClone(default_book),
            );
            const book2: BookData = await bookService.insert(
                structuredClone(default_book),
            );

            const result: BookDataResolved[] =
                await bookService.getByIdsResolved([book1.id, book2.id]);

            expect(result.length).toBe(2);
            for (let i = 0; i < result.length; i++) {
                expect(result[i].id).toBe(i === 0 ? book1.id : book2.id);
                expect(result[i].owner_name).toBe("test@book.de");
                expect(result[i].owner_type).toBe("user");
                expect(result[i].owner_id).toBe(user_ids[0]);
            }
        });

        it("should gracefully handle empty arrays", async () => {
            const result: BookDataResolved[] =
                await bookService.getByIdsResolved([]);
            expect(result.length).toBe(0);
        });

        it("should return an partial array if any of the books do not exist", async () => {
            const books = await bookService.getByIdsResolved([88888, 99999]);
            expect(books.length).toBe(0);

            const book1: BookData = await bookService.insert(
                structuredClone(default_book),
            );
            const books2 = await bookService.getByIdsResolved([
                book1.id,
                99999,
            ]);
            expect(books2.length).toBe(1);
            // Expect warn
            expect(consoleSpy).toHaveBeenCalled();
        });
    });

    describe("getAll", () => {
        it("should return all books in the database", async () => {
            const book1: BookData = await bookService.insert(
                structuredClone(default_book),
            );
            const book2: BookData = await bookService.insert(
                structuredClone(default_book),
            );

            const result: BookData[] = await bookService.getAll();

            expect(result.length).toBeGreaterThanOrEqual(2);
        });
    });

    describe("update", () => {
        it("should update the book in the database", async () => {
            const book: BookData = await bookService.insert(
                structuredClone(default_book),
            );

            const updatedBook = {
                ...book,
                title: "test3.book",
                comment: "This is an updated test book",
            };
            const result: BookData = await bookService.update(
                structuredClone(updatedBook),
            );

            expect_same(result, updatedBook, 10);

            //Check last_updated changed
            expect(result.last_updated!.getTime()).toBeGreaterThan(
                book.last_updated!.getTime(),
            );
        });

        it("should throw an error if the book does not exist", async () => {
            const book: BookData = await bookService.insert(
                structuredClone(default_book),
            );

            const updatedBook = {
                ...book,
                title: "test4.book",
                comment: "This is an updated test book",
                id: -1,
            };

            await expect(
                bookService.update(structuredClone(updatedBook)),
            ).rejects.toThrow();
        });
    });

    describe("delete", () => {
        it("should delete the book from the database", async () => {
            const book: BookData = await bookService.insert(
                structuredClone(default_book),
            );

            await bookService.delete(book.id);

            await expect(bookService.getById(book.id)).rejects.toThrow();
        });

        it("should throw an error if the book does not exist", async () => {
            await expect(bookService.delete(-1)).rejects.toThrow();
        });
    });

    describe("setOwner|setOwnerByName", () => {
        it("should set the owner of the book", async () => {
            const book: BookData = await bookService.insert(
                structuredClone(default_book),
            );

            ["user", ENTITY_USER].forEach(async (entity) => {
                const result = await bookService.setOwner(
                    user_ids[1],
                    entity as Entity,
                    book.id,
                );

                expect(result).toBe(true);
            });
            ["group", ENTITY_GROUP].forEach(async (entity) => {
                const result1 = await bookService.setOwner(
                    group_ids[0],
                    entity as Entity,
                    book.id,
                );
                expect(result1).toBe(true);
            });
        });

        it("should set the owner of the book by name", async () => {
            const book: BookData = await bookService.insert(
                structuredClone(default_book),
            );

            const result: BookOwner = await bookService.setOwnerByName(
                "test2@book.de",
                book.id,
            );

            expect(result.entity_name).toBe("test2@book.de");
            expect(result.entity_type).toBe("user");
            expect(result.id).toBe(user_ids[1]);
        });

        it("should throw an error if the book does not exist", async () => {
            await expect(
                bookService.setOwner(user_ids[1], ENTITY_USER, -1),
            ).rejects.toThrow();
        });

        it("should throw an error if the user does not exist", async () => {
            const book: BookData = await bookService.insert(
                structuredClone(default_book),
            );

            await expect(
                bookService.setOwner(-1, ENTITY_USER, book.id),
            ).rejects.toThrow();

            await expect(
                bookService.setOwnerByName("invalid", book.id),
            ).rejects.toThrow();
        });
    });

    describe("search", () => {
        test.todo("add test");
    });
});

function expect_same(
    book1: PartialBy<BookData, "id">,
    book2: PartialBy<BookData, "id">,
    time_diff: number = 4,
) {
    expect(book1.title).toBe(book2.title);
    expect(book1.comment).toBe(book2.comment);
    expect(book1.finished).toBe(book2.finished);
    expect(book1.owner_user_id).toBe(book2.owner_user_id);
    expect(book1.owner_group_id).toBe(book2.owner_group_id);
    expect(book1.cover_page_id).toBe(book2.cover_page_id);
    expect(book1.created!.getTime()).toBeCloseTo(
        book2.created!.getTime(),
        -time_diff,
    );
    expect(book1.last_updated!.getTime()).toBeCloseTo(
        book2.last_updated!.getTime(),
        -time_diff,
    );
}
