import { registerService } from "@snip/utils";

import { check_mariadb_warning, NotFoundError } from "../errors";
import { pool_permissions as pool } from "../sql.connection";
import { user_configData } from "../sql.typings";
import { UserConfigStrategy } from "../strategies";
import { ID, PartialBy, UserConfigData } from "../types";

export class UserConfigService implements UserConfigStrategy {
    /**
     * Retrieves all user configuration data from the database.
     * @returns A promise that resolves to an array of UserConfigData objects.
     */
    async getAll(): Promise<UserConfigData[]> {
        const sql = "SELECT * FROM user_config_resolved";
        const configData: UserConfigData[] = await pool
            .query(sql)
            .then(check_mariadb_warning)
            .then((res) => {
                return res;
            });

        return configData;
    }

    private async parseResolvedToNonResolved(
        data: Partial<UserConfigData>,
    ): Promise<Partial<user_configData>> {
        if (data.text_font !== undefined) {
            const font_id = await pool
                .query("SELECT id FROM fonts WHERE name = ?", [data.text_font])
                .then(check_mariadb_warning)
                .then((res) => {
                    if (res.length != 1) {
                        throw new Error(
                            "Font not found for name: " + data.text_font,
                        );
                    }
                    return res[0].id;
                });
            data.text_font_id = font_id;
        }

        const non_resolved: Partial<user_configData> = {
            // Overlap
            id: data.id,
        };
        for (const key in data) {
            if (key === "text_font") {
                continue;
            }
            non_resolved[key] = data[key];
        }

        return non_resolved;
    }

    // keys of configData
    allowedFields = [
        "pen_color",
        "pen_colors",
        "pen_size",
        "pen_smoothing",
        "text_font_id",
        "text_fontColor",
        "text_fontSize",
        "text_lineHeight",
        "text_lineWrap",
        "user_id",
        "zoom1",
        "zoom2",
    ];

    private parseToSqlFieldNamesAndValues(
        object: Partial<user_configData>,
    ): [unknown[], unknown[]] {
        if (Object.keys(object).length == 0) {
            throw new Error("No fields to update");
        }
        const field_names = Object.keys(object).filter((key) =>
            this.allowedFields.includes(key),
        );
        const field_values = field_names.map((key) => object[key]);

        //special case for pen_colors
        if (field_names.includes("pen_colors")) {
            field_values[field_names.indexOf("pen_colors")] = JSON.stringify(
                field_values[field_names.indexOf("pen_colors")],
            );
        }

        return [field_names, field_values];
    }

    async insert(
        data: PartialBy<UserConfigData, "id">,
    ): Promise<UserConfigData> {
        // Resolve text_font to text_font_id and create sql
        const non_resolved = await this.parseResolvedToNonResolved(data);
        const [field_names, vals] =
            this.parseToSqlFieldNamesAndValues(non_resolved);

        const sql =
            "INSERT INTO user_config (" +
            field_names.join(",") +
            ") VALUES (" +
            vals.map(() => "?").join(",") +
            ")";

        const configData_id = await pool
            .query(sql, vals)
            .then(check_mariadb_warning)
            .then((res) => {
                return res.insertId;
            });

        data.id = configData_id;
        data.last_updated = new Date();

        return data as UserConfigData;
    }

    /**
     * Updates a user configuration using partial data.
     * @param data The partial data to update, including the ID of the user configuration.
     * @returns A promise that resolves to the updated user configuration data.
     */
    async update(
        data: Partial<UserConfigData> & { id: ID },
    ): Promise<UserConfigData> {
        const non_resolved = await this.parseResolvedToNonResolved(data);
        const [field_names, vals] =
            this.parseToSqlFieldNamesAndValues(non_resolved);

        const sql = `UPDATE user_config SET ${field_names.map(
            (key) => `${key} = ?`,
        )} WHERE id = ?`;

        const configData = await pool
            .query(sql, [...vals, data.id])
            .then(check_mariadb_warning)
            .then((res) => {
                if (res.affectedRows != 1) {
                    throw new Error(
                        "Config not found for id:" + non_resolved.id,
                    );
                }
            })
            .then(() => {
                return this.getById(non_resolved.id);
            });

        return configData;
    }

    async delete(id: ID): Promise<boolean> {
        const sql = "DELETE FROM user_config WHERE id = ?";
        const configData = await pool
            .query(sql, [id])
            .then(check_mariadb_warning)
            .then((res) => {
                if (res.affectedRows != 1) {
                    throw new Error("Config not found for id:" + id);
                }
                return true;
            });

        return configData;
    }

    /**
     * Retrieves a user configuration by its ID.
     * @param id The ID of the user configuration.
     * @returns A promise that resolves to the user configuration data.
     */
    async getById(id: ID): Promise<UserConfigData> {
        const sql = "SELECT * FROM user_config_resolved WHERE id = ?";
        const configData = await pool
            .query(sql, [id])
            .then(check_mariadb_warning)
            .then((res) => {
                if (res.length != 1) {
                    throw new NotFoundError("user_config", id);
                }
                // As the database structure matches our UserConfigData interface
                return res[0];
            });

        return configData;
    }

    async getByIds(ids: ID[]): Promise<UserConfigData[]> {
        throw new Error("Method not implemented.");
    }

    /**
     * Retrieves the user configuration data by user ID.
     * If the user ID is not provided or is -1, it returns the default configuration.
     * If the configuration is not found for the given user ID, it throws an error.
     *
     * @param user_id - The ID of the user.
     * @returns A promise that resolves to the user configuration data.
     */
    async getByUserId(user_id: ID): Promise<UserConfigData> {
        const sql = "SELECT * FROM user_config_resolved WHERE user_id = ?";
        const configData = await pool
            .query(sql, [user_id])
            .then(check_mariadb_warning)
            .then((res) => {
                if (res.length != 1) {
                    throw new Error("Config not found for user_id: " + user_id);
                }
                // As the database structure matches our UserConfigData interface
                return res[0];
            });

        return configData;
    }

    /**
     * Updates the user configuration data for a specific user ID.
     *
     * @param user_id - The ID of the user.
     * @param configUpd - The partial user configuration data to update.
     * @returns A promise that resolves to a boolean indicating whether the update was successful.
     */
    async updateByUserId(
        data: Partial<UserConfigData> & { user_id: ID },
    ): Promise<UserConfigData> {
        // Remove user_id from data
        const user_id = data.user_id;
        delete data.user_id;

        // Params for insert
        const non_resolved = await this.parseResolvedToNonResolved(data);
        const [field_names, vals] =
            this.parseToSqlFieldNamesAndValues(non_resolved);

        const sql = `UPDATE user_config SET ${field_names.map(
            (key) => `${key} = ?`,
        )} WHERE user_id = ?`;

        const configData: UserConfigData = await pool
            .query(sql, [...vals, user_id])
            .then(check_mariadb_warning)
            .then((res) => {
                if (res.affectedRows != 1) {
                    throw new Error(
                        "Config not found for user_id: " + data.user_id,
                    );
                }
            })
            .then(() => {
                return this.getByUserId(user_id);
            });

        return configData;
    }
    async search(query: string): Promise<UserConfigData[]> {
        throw new Error("Method not implemented.");
    }
}

/** We need to globally register each service here
 * to prevent memory leaks in development.
 */
export const userConfigService = registerService(
    "db_userconfig",
    () => new UserConfigService(),
);
