import { BaseData, BaseView } from "@snip/database/types";
import { isValidHttpUrl } from "@snip/utils";

import { BaseSnip, BaseSnipArgs, EventContext, SnipData } from "./base";

export interface LinkData extends Omit<BaseData, "x" | "y"> {
    snip: SnipData<BaseData, BaseView>;
    href: string;
}

export interface LinkView extends BaseView {
    // Nothing here to see
}

export interface LinkSnipArgs extends Omit<BaseSnipArgs, "x" | "y"> {
    /** The underlying snip which gets gets an interactive link
     *
     *  default: null
     */
    snip: BaseSnip;

    /** An href to the linked resource
     *
     * default: null
     */
    href: string;
}

export class LinkSnip extends BaseSnip {
    public type = "link";

    public snip: BaseSnip;
    public href: URL;

    constructor({ snip, href, ...baseArgs }: LinkSnipArgs) {
        super({
            ...baseArgs,
        });

        this.snip = snip;

        // Check if the href is a valid URL
        if (!isValidHttpUrl(href)) {
            throw new Error("Invalid URL, must be a valid HTTP URL");
        }
        this.href = new URL(href);
    }

    /** Construct a instance of the snip from a
     * SnipData object. This is used to create a snip
     * object from the database data dict.
     *
     * @param {SnipData<ArrayData, ArrayView>} data - Data dict
     */
    static from_data(
        data: SnipData<LinkData, LinkView>,
        // eslint-disable-next-line @typescript-eslint/no-explicit-any
        get_snip_from_data?: any,
    ): LinkSnip {
        // Get snips from data
        const snip = get_snip_from_data(data.data.snip);

        return new LinkSnip({
            //Link
            snip,
            href: data.data.href,
            // Base
            page_id: data.page_id,
            book_id: data.book_id,
            last_updated: data.last_updated,
            created: data.created,
            //x,y is not needed
        });
    }

    /** Overload getters and setters to use
     * the inner snip as the link does not
     * have any coordinates.
     */
    render(ctx: CanvasRenderingContext2D): void {
        this.snip.render(ctx);
    }
    get x(): number {
        return this.snip.x;
    }
    set x(value: number) {
        this.snip.x = value;
    }
    get y(): number {
        return this.snip.y;
    }
    set y(value: number) {
        this.snip.y = value;
    }
    get height(): number {
        return this.snip.height;
    }
    get width(): number {
        return this.snip.width;
    }

    /** Handler for ui interactions, these are called
     * if the user clicks on the snip in the interaction mode
     * see base.ts for more explanations
     */
    on_dblclick = (context: EventContext) => {
        // Open link in new tab
        // TODO
        console.log("Open link in new tab", this.href);
        return;
    };
}
